# -*- coding: utf-8 -*-
"""
Created on Thu May 27 01:28:38 2021

@author: senam
"""

import os
import pandas as pd
import numpy as np

PENETRATION_STEP = 5
METRIC_MAP = {
    "thermal": {
        "submetrics": [
            "line_max_instantaneous_loading_pct",
            "line_max_moving_average_loading_pct",
            "line_num_time_points_with_instantaneous_violations",
            "line_num_time_points_with_moving_average_violations",
            "transformer_max_instantaneous_loading_pct",
            "transformer_max_moving_average_loading_pct",
            "transformer_num_time_points_with_instantaneous_violations",
            "transformer_num_time_points_with_moving_average_violations",
        ]
    },
    "voltage": {
        "node_type": ["primaries", "secondaries"],
        "submetrics": [
            "min_voltage",
            "max_voltage",
            "num_nodes_any_outside_ansi_b",
            "num_time_points_with_ansi_b_violations",
        ],
    }
}


def build_queries(columns, thresholds, metric_class, on="all"):
    """Build queries for filtering metrics dataframes"""
    queries = []
    if on == "all":
        on = METRIC_MAP[metric_class]["submetrics"]
    metrics = [m for m in on if m in columns]
    for metric in metrics:
        if "min" in metric:
            comparison = ">="
        else:
            comparison = "<="
        query = f"{metric} {comparison} {thresholds[metric_class][metric]}"
        queries.append(query)
    return queries


def synthesize_voltage(results_df):
    """ Reduce voltage metrics table to one time-point like table
    where for each metric, only the worst metric value of all time-points
    is recorded
    """
    filter_cols = ["name",
                   "substation",
                   "feeder",
                   "placement",
                   "sample",
                   "penetration_level",
                   "scenario",
    ]

    df = results_df.groupby(filter_cols)[["min_voltage"]].min().reset_index()
    df2 = results_df.groupby(filter_cols)[["max_voltage"]].max().reset_index()
    df = df.merge(df2, how="left", on=filter_cols)
    df3 = (
        results_df.groupby(filter_cols)[
            [
                "num_nodes_any_outside_ansi_b",
                "num_time_points_with_ansi_b_violations",
            ]
        ]
        .max()
        .reset_index()
    )

    df = df.merge(df3, how="left", on=filter_cols)

    return df

def synthesize_thermal(results_df):
    """ Reduce thermal metrics table to one time-point like table
    where for each metric, only the worst metric value of all time-points
    is recorded
    """

    filter_cols = ["name",
                   "substation",
                   "feeder",
                   "placement",
                   "sample",
                   "penetration_level",
                   "scenario"]
    df = (
        results_df.groupby(filter_cols)[
            [
                c for c in results_df.columns if (c not in filter_cols) and (c != "time_point")
                ]
        ]
        .max()
        .reset_index()
    )
    return df

def synthesize(metrics_df, metadata_df, metric_class):
    """Reduce metrics and metadata tables to one time-point like tables
    where for each metric, only the worst metric value of all time-points
    is recorded
    """
    if metric_class == 'voltage':
        # Both snapshot and time-series have primaries + secondaries.
        # Snapshot also has several time points.
        metrics_df = synthesize_voltage(metrics_df)

    # the presence of 'time_point' in the dataframe
    # indicates that we are dealing with a snapshot case"""
    elif metric_class == 'thermal' and 'time_point' in metrics_df.columns:
        metrics_df = synthesize_thermal(metrics_df)

    return metrics_df, metadata_df


def compute_hc_per_metric_class(
    result_path,
    thresholds,
    metric_class,
    scenario,
    node_types,
    on,
    hc_summary,
):
    """
    Given the metric class, compute its hosting capacity

    Parameters
    ----------
    result_path: str, the output directory of metrics summary tables
    thresholds: dict, the mapping of metric thresholds
    metric_class: str, the metric class, 'voltage' or 'thermal'
    node_types: list, the node types in voltage scenario
    on: list | str, the list of metrics of interest
        example: on = ['min_voltage', 'max_voltage']
    hc_summary: dict

    """
    metric_table = f"{metric_class}_metrics_table.csv"
    metric_df = pd.read_csv(
        os.path.join(result_path, metric_table),
        dtype={"sample": np.float64, "penetration_level": np.float64, "placement": str},
    )
    metric_df = metric_df.dropna(axis="index", subset=["sample", "penetration_level"])
    metric_df = metric_df[metric_df.scenario == scenario]
    meta_df = pd.read_csv(
        os.path.join(result_path, "metadata_table.csv"),
        dtype={"sample": np.float64, "penetration_level": np.float64, "placement": str},
    )
    meta_df = meta_df.dropna(axis="index", subset=["sample", "penetration_level"])

    if metric_class == "voltage" and len(node_types) == 1:
        metric_df = metric_df[metric_df.node_type == node_types[0]]

    metric_df, meta_df = synthesize(metric_df, meta_df, metric_class)

    queries = build_queries(metric_df.columns, thresholds, metric_class, on=on)
    query_phrase = " & ".join(queries)

    metric_df.penetration_level = metric_df.penetration_level.astype("float")
    if query_phrase:
        hc_summary = get_hosting_capacity(
            meta_df, metric_df, query_phrase, metric_class, hc_summary
        )

    if metric_class == "thermal":
        if (metric_df.transformer_instantaneous_threshold.isna().any() or 
            metric_df.transformer_instantaneous_threshold.isnull().any()):
            queries = [q for q in queries if 'transformer' not in q]
            noxfmr_metric_df = metric_df.loc[metric_df.transformer_instantaneous_threshold.isna(), :]
            noxfmr_query_phrase = " & ".join(queries)
            if noxfmr_query_phrase:
                hc_summary = get_hosting_capacity(
                    meta_df, noxfmr_metric_df, noxfmr_query_phrase, metric_class, hc_summary
                )
    
    return hc_summary, query_phrase


def get_hosting_capacity(meta_df, metric_df, query_phrase, metric_class, hc_summary):
    """Return the hosting capacity summary

    violation_starting_penetration: the lowest penetration level at which the analysis revealed a violation
    candidate_cba_samples: list of all deployments or samples that have violations at the lowest penetration level
    violation_frequency_by_sample: the ratio between the number of scenarios yielding violations and the total number of scenarios investigated in a given deployment sample
    recommended_cba_sample: the sample with the highest violation frequency, recommended for further cost benefit analysis or upgrade

    """
    
    cba_samples = set(metric_df['sample'])
    violation_starting_penetration = 0
    violation_frequency_by_sample = {}

    for feeder, temp_df in metric_df.groupby(by="feeder"):
        if not feeder in hc_summary.keys():
            hc_summary[feeder] = dict()
        temp_pass = temp_df.query(query_phrase)
        temp_fail = temp_df.query(f"~({query_phrase})")
        violation_frequency_by_sample = {d:len(temp_fail.query('sample == @d'))/len(temp_df.query('sample == @d')) for d in temp_df['sample'].unique()}
        recommended_cba_sample = max(violation_frequency_by_sample, key=violation_frequency_by_sample.get)
        fail_penetration_levels = set(temp_fail.penetration_level.values)
        pass_penetration_levels = set(temp_pass.penetration_level.values)
        if fail_penetration_levels:
            violation_starting_penetration = min(fail_penetration_levels)
            cba_samples = set(temp_fail.loc[temp_fail.penetration_level==violation_starting_penetration, 'sample'])
        else:
            violation_starting_penetration = None
            cba_samples = set()
        temp_min_values = pass_penetration_levels.difference(fail_penetration_levels)
        if len(temp_pass) == 0:
            # min_hc = min(temp_fail.penetration_level.values)
            min_hc = 0 # This is supposed to be the PV penetration level of the base case if it passed, 0 otherwise
            max_hc = 0 # This is supposed to be the PV penetration level of the base case if it passed, 0 otherwise
        elif len(temp_fail) == 0:
            min_hc = max(pass_penetration_levels)
            max_hc = max(pass_penetration_levels)
        else:
            max_hc = max(pass_penetration_levels)
            if temp_min_values:
                min_hc = max(temp_min_values)
            else:
                min_hc = 0

        total_feeder_load = meta_df[meta_df.feeder == feeder][
            "load_capacity_kw"
        ].values[0]
        max_kW = max_hc * total_feeder_load / 100
        min_kW = min_hc * total_feeder_load / 100

        hc_summary[feeder][metric_class] = {
            "min_hc_pct": min_hc,
            "max_hc_pct": max_hc,
            "min_hc_kw": round(min_kW, 0),
            "max_hc_kw": round(max_kW, 0),
            "violation_starting_penetration": violation_starting_penetration,
            "candidate_cba_samples": list(cba_samples),
            "violation_frequency_by_sample": violation_frequency_by_sample,
            "recommended_cba_sample": recommended_cba_sample,
        }
    return hc_summary


def compute_hc(
    result_path,
    thresholds,
    metric_classes,
    scenario,
    node_types,
    on="all",
):
    """
    Compute hosting capacity

    Parameters
    ----------
    result_path: str, the output directory of metrics summary tables
    thresholds: dict, the mapping of metric thresholds
    metric_classes: list, the list of metric class
    node_types: list, the node types in voltage scenario
    on: list | str, the list of metrics of interest
        example: on = ['min_voltage', 'max_voltage']

    Returns
    -------
    dict: hc_summary
    dict: hc_overall
    list: query strings
    """
    query_list = []
    hc_summary = {}
    hc_overall = {}
    for metric_class in metric_classes:
        hc_summary, query_phrase = compute_hc_per_metric_class(
            result_path,
            thresholds,
            metric_class,
            scenario,
            node_types,
            on,
            hc_summary,
        )
        query_list.append(query_phrase)
    for feeder, dic in hc_summary.items():
        hc_overall[feeder] = {}
        df = pd.DataFrame.from_dict(dic, "index")
        for column in df.columns:
            if 'hc' in column:
                hc_overall[feeder][column] = min(df[column])
        th_sample = dic['thermal']['recommended_cba_sample']
        th_samples = dic['thermal']['candidate_cba_samples']
        v_sample = dic['voltage']['recommended_cba_sample']
        v_samples = dic['voltage']['candidate_cba_samples']
        cba_rec_pen = list(np.arange(
		    hc_overall[feeder]['min_hc_pct'] + PENETRATION_STEP, 
            hc_overall[feeder]['max_hc_pct'] + PENETRATION_STEP, 
            PENETRATION_STEP)
		)
        if th_sample in v_samples:
            hc_overall[feeder]['cba_recommendation'] = [
			    {'sample': th_sample, 'penetrations': cba_rec_pen}
			]
        elif v_sample in th_samples:
            hc_overall[feeder]['cba_recommendation'] = [
			    {'sample':v_sample, 'penetrations':cba_rec_pen}
			]
        else:
            v_cba_rec_pen = list(np.arange(
			    dic['voltage']['min_hc_pct'] + PENETRATION_STEP, 
                dic['voltage']['max_hc_pct'] + PENETRATION_STEP, 
                PENETRATION_STEP
            ))
            th_cba_rec_pen = list(np.arange(
			    dic['thermal']['min_hc_pct'] + PENETRATION_STEP, 
                dic['thermal']['max_hc_pct'] + PENETRATION_STEP, 
                PENETRATION_STEP
            ))
            hc_overall[feeder]['cba_recommendation'] = [
                {'sample': v_sample, 'penetrations': v_cba_rec_pen},
                {'sample': th_sample, 'penetrations': th_cba_rec_pen}
            ]

    return hc_summary, hc_overall, query_list



import logging
import os

import click

from jade.loggers import setup_logging
from jade.utils.utils import load_data

from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO

from disco.cli.config_snapshot import (
    switch_snapshot_to_qsts,
    common_snapshot_options,
    make_simulation_config,
)
from disco.cli.config_time_series import common_time_series_options
from disco.models.base import OpenDssDeploymentModel, SimulationModel
from disco.models.power_flow_generic_models import (
    PowerFlowSnapshotSimulationModel,
    PowerFlowTimeSeriesSimulationModel,
)
from disco.distribution.deployment_parameters import DeploymentParameters
from disco.enums import SimulationType
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.pydss.common import ConfigType


logger = logging.getLogger(__name__)


@click.group()
def config_generic_models():
    """Create a JADE config file from a set of generic OpenDSS models."""


@click.command()
@click.argument("power-flow-config-file", type=click.Path(exists=True))
@common_snapshot_options
def snapshot(
    power_flow_config_file,
    config_file,
    exports_filename,
    reports_filename,
    with_loadshape,
    auto_select_time_points,
    auto_select_time_points_search_duration_days,
    shuffle,
    store_per_element_data,
    strip_whitespace,
    volt_var_curve,
    verbose,
):
    """Create a JADE config file for a snapshot power-flow simulation."""
    pf_config = PowerFlowSnapshotSimulationModel.from_file(power_flow_config_file)
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])

    simulation_config, scenarios = make_simulation_config(
        reports_filename,
        exports_filename,
        pf_config.include_pf1,
        pf_config.include_control_mode,
        store_per_element_data,
        with_loadshape,
        auto_select_time_points,
        auto_select_time_points_search_duration_days,
    )
    config = PyDssConfiguration()
    simulation_type = SimulationType.QSTS if with_loadshape else SimulationType.SNAPSHOT
    for job in pf_config.jobs:
        job = DeploymentParameters(
            model_type="SnapshotImpactAnalysisModel",
            name=job.name,
            blocked_by=job.blocked_by,
            base_case=None,
            is_base_case=False,
            estimated_run_minutes=job.estimated_run_minutes,
            deployment=OpenDssDeploymentModel(
                is_standalone=True,
                deployment_file=job.opendss_model_file,
                substation=job.substation or "NA",
                feeder=job.feeder or "NA",
                dc_ac_ratio=1.0,
                directory=os.path.dirname(job.opendss_model_file),
                kva_to_kw_rating=1.0,
                project_data=job.project_data,
                pydss_controllers=job.pydss_controllers,
            ),
            simulation=SimulationModel(
                start_time=pf_config.start_time,
                end_time=pf_config.start_time,
                step_resolution=900,
                simulation_type=simulation_type,
            ),
        )
        config.add_job(job)

    config.check_job_consistency()

    if volt_var_curve is not None:
        if config.has_pydss_controllers() and pf_config.include_control_mode:
            config.update_volt_var_curve(volt_var_curve)
        else:
            logger.warning(
                "Setting a volt_var_curve has no effect when there is no %s scenario.",
                CONTROL_MODE_SCENARIO,
            )

    config.set_pydss_config(ConfigType.SIMULATION_CONFIG, simulation_config)
    config.set_pydss_config(ConfigType.SCENARIOS, scenarios)

    # We can't currently predict how long each job will take. If we did, we could set
    # estimated_run_minutes for each job.
    # Shuffle the jobs randomly so that we have a better chance of getting batches with similar
    # runtimes.
    if shuffle:
        config.shuffle_jobs()

    if with_loadshape:
        config = switch_snapshot_to_qsts(config)

    indent = None if strip_whitespace else 2
    config.dump(filename=config_file, indent=indent)
    print(f"Created {config_file} for Snapshot Analysis")


@click.command()
@click.argument("power-flow-config-file", type=click.Path(exists=True))
@common_time_series_options
def time_series(
    power_flow_config_file,
    config_file,
    feeder_losses,
    pv_clipping,
    pv_curtailment,
    thermal_metrics,
    voltage_metrics,
    capacitor_changes,
    regcontrol_changes,
    export_data_tables,
    exports_filename,
    reports_filename,
    skip_night,
    store_all_time_points,
    store_per_element_data,
    volt_var_curve,
    verbose,
):
    """Create JADE configuration for time series simulations."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])

    pf_config = PowerFlowTimeSeriesSimulationModel.from_file(power_flow_config_file)
    simulation_config = PyDssConfiguration.get_default_pydss_simulation_config()
    simulation_config["project"]["simulation_type"] = SimulationType.QSTS.value
    simulation_config["reports"] = load_data(reports_filename)["reports"]
    simulation_config["exports"]["export_data_tables"] = export_data_tables
    for report in simulation_config["reports"]["types"]:
        if report["name"] == "Feeder Losses" and feeder_losses is not None:
            report["enabled"] = feeder_losses
        if report["name"] == "PV Clipping" and pv_clipping is not None:
            report["enabled"] = pv_clipping
        if report["name"] == "PV Curtailment" and pv_curtailment is not None:
            report["enabled"] = pv_curtailment
        if report["name"] == "Thermal Metrics" and thermal_metrics is not None:
            report["enabled"] = thermal_metrics
            report["store_all_time_points"] = store_all_time_points
        if report["name"] == "Voltage Metrics" and voltage_metrics is not None:
            report["enabled"] = voltage_metrics
        if report["name"] in ("Thermal Metrics", "Voltage Metrics"):
            report["store_per_element_data"] = store_per_element_data
            report["store_all_time_points"] = store_all_time_points
        if report["name"] == "Capacitor State Change Counts" and capacitor_changes is not None:
            report["enabled"] = capacitor_changes
        if report["name"] == "RegControl Tap Number Change Counts" and regcontrol_changes is not None:
            report["enabled"] = regcontrol_changes

    exports = {} if exports_filename is None else load_data(exports_filename)
    scenarios = []
    if pf_config.include_control_mode:
        scenarios.append(
            PyDssConfiguration.make_default_pydss_scenario(CONTROL_MODE_SCENARIO, exports),
        )
    if pf_config.include_pf1:
        scenarios.append(
            PyDssConfiguration.make_default_pydss_scenario(PF1_SCENARIO, exports),
        )

    config = PyDssConfiguration()
    for job in pf_config.jobs:
        job = DeploymentParameters(
            model_type="TimeSeriesAnalysisModel",
            name=job.name,
            blocked_by=job.blocked_by,
            base_case=None,
            is_base_case=False,
            estimated_run_minutes=job.estimated_run_minutes,
            deployment=OpenDssDeploymentModel(
                is_standalone=True,
                deployment_file=job.opendss_model_file,
                substation=job.substation or "NA",
                feeder=job.feeder or "NA",
                dc_ac_ratio=1.0,
                directory=os.path.dirname(job.opendss_model_file),
                kva_to_kw_rating=1.0,
                project_data=job.project_data,
                pydss_controllers=job.pydss_controllers,
            ),
            simulation=SimulationModel(
                start_time=pf_config.start_time,
                end_time=pf_config.end_time,
                step_resolution=pf_config.step_resolution,
                simulation_type=SimulationType.QSTS,
            ),
        )
        config.add_job(job)

    config.check_job_consistency()

    if volt_var_curve is not None:
        if config.has_pydss_controllers() and pf_config.include_control_mode:
            config.update_volt_var_curve(volt_var_curve)
        else:
            logger.warning(
                "Setting a volt_var_curve has no effect when there is no %s scenario.",
                CONTROL_MODE_SCENARIO,
            )

    if skip_night:
        pydss_sim_config = config.get_pydss_config(ConfigType.SIMULATION_CONFIG)
        pydss_sim_config["project"]["simulation_range"] = {"start": "06:00:00", "end": "18:00:00"}
        # Note that we are using the same convergence error threshold percent.
        config.set_pydss_config(ConfigType.SIMULATION_CONFIG, pydss_sim_config)

    config.set_pydss_config(ConfigType.SIMULATION_CONFIG, simulation_config)
    config.set_pydss_config(ConfigType.SCENARIOS, scenarios)
    config.dump(filename=config_file)

    print(f"Created {config_file} for TimeSeries Analysis")


config_generic_models.add_command(snapshot)
config_generic_models.add_command(time_series)



import time
import json
import logging
import pandas as pd

from jade.utils.timing_utils import track_timing, Timer
from jade.utils.utils import load_data, dump_data

from .thermal_upgrade_functions import *
from .voltage_upgrade_functions import plot_thermal_violations, plot_voltage_violations, plot_feeder

from disco.models.upgrade_cost_analysis_generic_input_model import UpgradeTechnicalCatalogModel
from disco.models.upgrade_cost_analysis_generic_output_model import UpgradeViolationResultModel, AllUpgradesTechnicalResultModel
from disco import timer_stats_collector
from disco.enums import LoadMultiplierType
from disco.exceptions import UpgradesInvalidViolationIncrease


logger = logging.getLogger(__name__)


@track_timing(timer_stats_collector)
def determine_thermal_upgrades(
    job_name,
    master_path,
    enable_pydss_solve,
    pydss_volt_var_model,
    thermal_config,
    internal_upgrades_technical_catalog_filepath,
    thermal_upgrades_dss_filepath,
    upgraded_master_dss_filepath,
    output_json_thermal_upgrades_filepath,
    feeder_stats_json_file,
    thermal_upgrades_directory,
    overall_output_summary_filepath,
    dc_ac_ratio,
    ignore_switch=True,
    verbose=False
):
    start_time = time.time()
    logger.info( f"Simulation start time: {start_time}")   
    initial_simulation_params = {"enable_pydss_solve": enable_pydss_solve, "pydss_volt_var_model": pydss_volt_var_model,
                                 "dc_ac_ratio": dc_ac_ratio}
    logger.info("Initial simulation parameters: %s", initial_simulation_params)
    create_plots = thermal_config["create_plots"]
    # start upgrades
    initial_dss_file_list = [master_path]
    simulation_params = reload_dss_circuit(dss_file_list=initial_dss_file_list, commands_list=None, **initial_simulation_params)
    timepoint_multipliers = thermal_config["timepoint_multipliers"]

    if timepoint_multipliers is not None:
        multiplier_type = LoadMultiplierType.UNIFORM
    else:
        multiplier_type = LoadMultiplierType.ORIGINAL
    simulation_params.update({"timepoint_multipliers": timepoint_multipliers, "multiplier_type": multiplier_type})
    
    voltage_upper_limit = thermal_config["voltage_upper_limit"]
    voltage_lower_limit = thermal_config["voltage_lower_limit"]
    if thermal_config["read_external_catalog"]:
        with open(thermal_config["external_catalog"]) as json_file:
            external_upgrades_technical_catalog = json.load(json_file)
        # perform validation for external catalog
        input_catalog_model = UpgradeTechnicalCatalogModel(**external_upgrades_technical_catalog)
        line_upgrade_options = pd.DataFrame.from_dict(input_catalog_model.dict(by_alias=True)["line"])
        xfmr_upgrade_options = pd.DataFrame.from_dict(input_catalog_model.dict(by_alias=True)["transformer"])
        # this will remove any duplicates if present
        line_upgrade_options = determine_available_line_upgrades(line_upgrade_options)
        xfmr_upgrade_options = determine_available_xfmr_upgrades(xfmr_upgrade_options)
    else:
        external_upgrades_technical_catalog = {}
        orig_lines_df = get_thermal_equipment_info(compute_loading=False, equipment_type="line")
        orig_xfmrs_df = get_thermal_equipment_info(compute_loading=False, equipment_type="transformer")
        orig_linecode_df = get_line_code()
        orig_linegeometry_df = get_line_geometry()
        line_upgrade_options = determine_available_line_upgrades(orig_lines_df)
        xfmr_upgrade_options = determine_available_xfmr_upgrades(orig_xfmrs_df)
        internal_upgrades_technical_catalog = {"line": line_upgrade_options.to_dict('records'), "transformer": xfmr_upgrade_options.to_dict('records'),
                                               "linecode": orig_linecode_df.to_dict('records'), 
                                               "geometry": orig_linegeometry_df.to_dict('records'),
                                               }
        # validate internal upgrades catalog
        input_catalog_model = UpgradeTechnicalCatalogModel(**internal_upgrades_technical_catalog)
        dump_data(input_catalog_model.dict(by_alias=True), 
                  internal_upgrades_technical_catalog_filepath, indent=2)  # write internal catalog to json
        # reassign from model to dataframes, so datatypes are maintained
        line_upgrade_options = pd.DataFrame.from_dict(input_catalog_model.dict(by_alias=True)["line"])
        xfmr_upgrade_options = pd.DataFrame.from_dict(input_catalog_model.dict(by_alias=True)["transformer"])
    # get these feeder details before running powerflow
    feeder_stats = {"feeder_metadata": {}, "stage_results": []}
    feeder_stats["feeder_metadata"].update(get_feeder_stats(dss))
    (
        initial_bus_voltages_df,
        initial_undervoltage_bus_list,
        initial_overvoltage_bus_list,
        initial_buses_with_violations,
    ) = get_bus_voltages(voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)
    
    initial_xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                         equipment_type="transformer", **simulation_params)
    initial_line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                         equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
    
    initial_overloaded_xfmr_list = list(initial_xfmr_loading_df.loc[initial_xfmr_loading_df["status"] == 
                                                                    "overloaded"]["name"].unique())
    initial_overloaded_line_list = list(initial_line_loading_df.loc[initial_line_loading_df["status"] == 
                                                                    "overloaded"]["name"].unique())
    orig_ckt_info = get_circuit_info()
    circuit_source = orig_ckt_info['source_bus']
    orig_regcontrols_df = get_regcontrol_info(correct_PT_ratio=False)
    orig_capacitors_df = get_capacitor_info(correct_PT_ratio=False)
    feeder_stats["stage_results"].append(get_upgrade_stage_stats(dss, upgrade_stage="initial", upgrade_type="thermal", xfmr_loading_df=initial_xfmr_loading_df, line_loading_df=initial_line_loading_df, 
                                        bus_voltages_df=initial_bus_voltages_df, capacitors_df=orig_capacitors_df, regcontrols_df=orig_regcontrols_df) )
    dump_data(feeder_stats, feeder_stats_json_file, indent=2)   # save feeder stats
    if len(initial_overloaded_xfmr_list) > 0 or len(initial_overloaded_line_list) > 0:
        n = len(initial_overloaded_xfmr_list) +  len(initial_overloaded_line_list)
        equipment_with_violations = {"Transformer": initial_xfmr_loading_df, "Line": initial_line_loading_df}
        if create_plots:
            plot_thermal_violations(fig_folder=thermal_upgrades_directory, title="Thermal violations before thermal upgrades_"+str(n), 
                                    equipment_with_violations=equipment_with_violations, circuit_source=circuit_source)
            plot_voltage_violations(fig_folder=thermal_upgrades_directory, title="Bus violations before thermal upgrades_"+str(len(initial_buses_with_violations)), 
                                    buses_with_violations=initial_buses_with_violations, circuit_source=circuit_source, enable_detailed=True)
        upgrade_status = "Thermal Upgrades Required"  # status - whether upgrades done or not
    else:
        upgrade_status = "Thermal Upgrades not Required"  # status - whether upgrades done or not
    logger.info(upgrade_status)
    scenario = get_scenario_name(enable_pydss_solve, pydss_volt_var_model)
    initial_results = UpgradeViolationResultModel(
        name=job_name,
        scenario=scenario,
        stage="initial",
        upgrade_type="thermal",
        # upgrade_status = upgrade_status,
        simulation_time_s = 0.0,
        thermal_violations_present=(len(initial_overloaded_xfmr_list) + len(initial_overloaded_line_list)) > 0,
        voltage_violations_present=(len(initial_undervoltage_bus_list) + len(initial_overvoltage_bus_list)) > 0,
        max_bus_voltage=initial_bus_voltages_df['max_per_unit_voltage'].max(),
        min_bus_voltage=initial_bus_voltages_df['min_per_unit_voltage'].min(),
        num_voltage_violation_buses=len(initial_undervoltage_bus_list) + len(initial_overvoltage_bus_list),
        num_overvoltage_violation_buses=len(initial_overvoltage_bus_list),
        voltage_upper_limit=voltage_upper_limit,
        num_undervoltage_violation_buses=len(initial_undervoltage_bus_list),
        voltage_lower_limit=voltage_lower_limit,
        max_line_loading=initial_line_loading_df['max_per_unit_loading'].max(),
        max_transformer_loading=initial_xfmr_loading_df['max_per_unit_loading'].max(),
        num_line_violations=len(initial_overloaded_line_list),
        line_upper_limit=thermal_config['line_upper_limit'],
        num_transformer_violations=len(initial_overloaded_xfmr_list),
        transformer_upper_limit=thermal_config['transformer_upper_limit']
    )
    temp_results = convert_dict_nan_to_none(dict(initial_results))
    if os.path.exists(overall_output_summary_filepath):
        overall_outputs = load_data(overall_output_summary_filepath)
        overall_outputs["violation_summary"].append(temp_results)
    else:
        overall_outputs = {"violation_summary": [temp_results]}
    dump_data(overall_outputs, overall_output_summary_filepath, indent=2, allow_nan=False)
    title = "Feeder"
    plot_feeder(fig_folder=thermal_upgrades_directory, title=title, circuit_source=circuit_source, enable_detailed=True)
    # Mitigate thermal violations
    iteration_counter = 0
    # if number of violations is very high,  limit it to a small number
    max_upgrade_iteration = min(
        thermal_config["upgrade_iteration_threshold"],
        len(initial_overloaded_xfmr_list) + len(initial_overloaded_line_list),)
    logger.info(f"Maximum allowable number of thermal upgrade iterations: {max_upgrade_iteration}")
    commands_list = []
    line_upgrades_df = pd.DataFrame()
    xfmr_upgrades_df = pd.DataFrame()
    overloaded_line_list = initial_overloaded_line_list
    overloaded_xfmr_list = initial_overloaded_xfmr_list
    while (len(overloaded_line_list) > 0 or len(overloaded_xfmr_list) > 0) and (
        iteration_counter < max_upgrade_iteration):
        line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                    equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
        overloaded_line_list = list(line_loading_df.loc[line_loading_df["status"] == "overloaded"]["name"].unique())
        logger.info(f"Iteration_{iteration_counter}: Determined line loadings.")
        logger.info(f"Iteration_{iteration_counter}: Number of line violations: {len(overloaded_line_list)}")
        before_upgrade_num_line_violations = len(overloaded_line_list)
        if len(overloaded_line_list) > 0:            
            line_commands_list, temp_line_upgrades_df = correct_line_violations(
                line_loading_df=line_loading_df,
                line_design_pu=thermal_config["line_design_pu"],
                line_upgrade_options=line_upgrade_options.copy(deep=True),
                parallel_lines_limit=thermal_config["parallel_lines_limit"],
                external_upgrades_technical_catalog=external_upgrades_technical_catalog,)
            logger.info(f"Iteration_{iteration_counter}: Corrected line violations.")
            commands_list = commands_list + line_commands_list
            line_upgrades_df = pd.concat([line_upgrades_df, temp_line_upgrades_df])
        xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                    equipment_type="transformer", **simulation_params)
        overloaded_xfmr_list = list(xfmr_loading_df.loc[xfmr_loading_df["status"] == "overloaded"]["name"].unique())
        logger.info(f"Iteration_{iteration_counter}: Determined xfmr loadings.")
        logger.info(f"Iteration_{iteration_counter}: Number of xfmr violations: {len(overloaded_xfmr_list)}")
        before_upgrade_num_xfmr_violations = len(overloaded_xfmr_list)
        
        if len(overloaded_xfmr_list) > 0:
            xfmr_commands_list, temp_xfmr_upgrades_df = correct_xfmr_violations(
                xfmr_loading_df=xfmr_loading_df,
                xfmr_design_pu=thermal_config["transformer_design_pu"],
                xfmr_upgrade_options=xfmr_upgrade_options.copy(deep=True),
                parallel_transformers_limit=thermal_config["parallel_transformers_limit"])
            logger.info(f"Iteration_{iteration_counter}: Corrected xfmr violations.")
            commands_list = commands_list + xfmr_commands_list
            xfmr_upgrades_df = pd.concat([xfmr_upgrades_df, temp_xfmr_upgrades_df])
        # compute loading after upgrades
        xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                    equipment_type="transformer",  **simulation_params)
        overloaded_xfmr_list = list(xfmr_loading_df.loc[xfmr_loading_df["status"] == "overloaded"]["name"].unique())
        line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                    equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
        overloaded_line_list = list(line_loading_df.loc[line_loading_df["status"] == "overloaded"]["name"].unique())
        
        if len(overloaded_line_list) > before_upgrade_num_line_violations:
            logger.debug(overloaded_line_list)
            logger.info("Write upgrades till this step to debug upgrades")
            write_text_file(string_list=commands_list, text_file_path=thermal_upgrades_dss_filepath)
            raise UpgradesInvalidViolationIncrease(f"Line violations increased from {before_upgrade_num_line_violations} to {len(overloaded_line_list)} "
                f"during upgrade process")
        if len(overloaded_xfmr_list) > before_upgrade_num_xfmr_violations:
            logger.info("Write upgrades till this step to debug upgrades")
            write_text_file(string_list=commands_list, text_file_path=thermal_upgrades_dss_filepath)
            raise UpgradesInvalidViolationIncrease(f"Xfmr violations increased from {before_upgrade_num_xfmr_violations} to {len(overloaded_xfmr_list)} "
                f"during upgrade process")

        logger.info(f"Iteration_{iteration_counter}: Number of devices with violations after this iteration: Transformers:{len(overloaded_xfmr_list)}, Lines: {len(overloaded_line_list)}")
        iteration_counter += 1
        if iteration_counter > max_upgrade_iteration:
            logger.info(f"Max iterations limit reached, quitting algorithm. This means all thermal violations were not resolved with these limited iterations."
                        f"You can increase the Iteration limit in the thermal_config['upgrade_iteration_threshold']")
            break
    if iteration_counter > 1:
        logger.info(f"Multiple iterations ({iteration_counter}) were needed to resolve thermal violations."
                    f"This indicates that feeder was extremely overloaded to start with.")
    
    if any("new " in string.lower() for string in commands_list):  # if new equipment is added.
        commands_list.append("CalcVoltageBases")
    commands_list.append("Solve")
    write_text_file(string_list=commands_list, text_file_path=thermal_upgrades_dss_filepath)
    redirect_command_list = create_upgraded_master_dss(dss_file_list=initial_dss_file_list + [thermal_upgrades_dss_filepath], upgraded_master_dss_filepath=upgraded_master_dss_filepath,
                                                       original_master_filename=os.path.basename(master_path))
    write_text_file(string_list=redirect_command_list, text_file_path=upgraded_master_dss_filepath)
    reload_dss_circuit(dss_file_list=[upgraded_master_dss_filepath], commands_list=None, **simulation_params,)
    bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(voltage_upper_limit=voltage_upper_limit, 
                                                                                                           voltage_lower_limit=voltage_lower_limit, **simulation_params)
    xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                equipment_type="transformer", **simulation_params)
    line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
    overloaded_xfmr_list = list(xfmr_loading_df.loc[xfmr_loading_df["status"] == "overloaded"]["name"].unique())
    overloaded_line_list = list(line_loading_df.loc[line_loading_df["status"] == "overloaded"]["name"].unique())
    # same equipment could be upgraded(edited) multiple times. Only consider highest capacity upgrade done. original_equipment details are currently not used.
    line_upgrades_df = remove_duplicate_line_upgrades(line_upgrades_df)
    xfmr_upgrades_df = remove_duplicate_transformer_upgrades(xfmr_upgrades_df)
    # validate upgrades output models
    m = AllUpgradesTechnicalResultModel(line=line_upgrades_df.to_dict('records'), transformer=xfmr_upgrades_df.to_dict('records'))
    temp = m.dict(by_alias=True)
    temp.pop("voltage")
    dump_data(temp, output_json_thermal_upgrades_filepath, indent=2)
    n = len(overloaded_xfmr_list) +  len(overloaded_line_list)
    equipment_with_violations = {"Transformer": xfmr_loading_df, "Line": line_loading_df}
    if (upgrade_status == "Thermal Upgrades Required") and create_plots:
        plot_thermal_violations(fig_folder=thermal_upgrades_directory, title="Thermal violations after thermal upgrades_"+str(n), 
                                equipment_with_violations=equipment_with_violations, circuit_source=circuit_source)
        plot_voltage_violations(fig_folder=thermal_upgrades_directory, title="Bus violations after thermal upgrades_"+str(len(buses_with_violations)), 
                                buses_with_violations=buses_with_violations, circuit_source=circuit_source, enable_detailed=True)
    if os.path.exists(feeder_stats_json_file):
        feeder_stats = load_data(feeder_stats_json_file)
    else:
        feeder_stats = {}
    regcontrols_df = get_regcontrol_info(correct_PT_ratio=False)
    capacitors_df = get_capacitor_info(correct_PT_ratio=False)
    feeder_stats["stage_results"].append( get_upgrade_stage_stats(dss, upgrade_stage="final", upgrade_type="thermal", xfmr_loading_df=xfmr_loading_df, line_loading_df=line_loading_df, 
                                        bus_voltages_df=bus_voltages_df, capacitors_df=capacitors_df, regcontrols_df=regcontrols_df) )
    dump_data(feeder_stats, feeder_stats_json_file, indent=2)
    end_time = time.time()
    logger.info(f"Simulation end time: {end_time}")
    simulation_time = end_time - start_time
    final_results = UpgradeViolationResultModel(
        name=job_name,
        scenario=scenario,
        stage="final",
        upgrade_type="thermal",
        # upgrade_status=upgrade_status,
        simulation_time_s=simulation_time,
        thermal_violations_present=(len(overloaded_xfmr_list) + len(overloaded_line_list)) > 0,
        voltage_violations_present=(len(undervoltage_bus_list) + len(overvoltage_bus_list)) > 0,
        max_bus_voltage=bus_voltages_df['max_per_unit_voltage'].max(),
        min_bus_voltage=bus_voltages_df['min_per_unit_voltage'].min(),
        num_voltage_violation_buses=len(undervoltage_bus_list) + len(overvoltage_bus_list),
        num_overvoltage_violation_buses=len(overvoltage_bus_list),
        voltage_upper_limit=voltage_upper_limit,
        num_undervoltage_violation_buses=len(undervoltage_bus_list),
        voltage_lower_limit=voltage_lower_limit,
        max_line_loading=line_loading_df['max_per_unit_loading'].max(),
        max_transformer_loading=xfmr_loading_df['max_per_unit_loading'].max(),
        num_line_violations=len(overloaded_line_list),
        line_upper_limit=thermal_config['line_upper_limit'],
        num_transformer_violations=len(overloaded_xfmr_list),
        transformer_upper_limit=thermal_config['transformer_upper_limit'],
    )
    temp_results = dict(final_results)
    temp_results = convert_dict_nan_to_none(temp_results)
    if os.path.exists(overall_output_summary_filepath):
        overall_outputs = load_data(overall_output_summary_filepath)
        overall_outputs["violation_summary"].append(temp_results)
    else:
        overall_outputs = {"violation_summary": [temp_results]}
    dump_data(overall_outputs, overall_output_summary_filepath, indent=2, allow_nan=False)



import logging
import os
import platform
import pathlib
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum

from disco.storage.exceptions import IngestionError

# NOTE: Important table names, please do not change the order.
TABLE_NAMES = [
    "feeder_head_table.csv",
    "feeder_losses_table.csv",
    "metadata_table.csv",
    "thermal_metrics_table.csv",
    "voltage_metrics_table.csv",
    
    # Only apply to snapshot simulaiton 
    # with --auto-select-time-points option enabled
    "snapshot_time_points_table.csv"
]

DERMS_INFO_FILENAME = "derms_simulation_info.json"

logger = logging.getLogger(__name__)


class OutputType(Enum):
    JADE = "jade"
    JADE_PIPELINE = "jade-pipeline"
    DERMS = "derms"


class OutputBase(ABC):
    """Abstract output class for handling JADE output directory"""

    def __init__(self, output: pathlib.Path) -> None:
        self.output = self.validate_output(output)

    @property
    @abstractmethod
    def output_type(self):
        """The output type"""

    @property
    def creation_time(self):
        timestamp = get_creation_time(self.output)
        return datetime.fromtimestamp(timestamp)

    @property
    def config_file(self):
        return self.output / "config.json"

    @property
    def result_file(self):
       return self.output / "results.json"

    @property
    def job_outputs(self):
        return self.output / "job-outputs"
    
    @property
    def table_names(self):
        return TABLE_NAMES
    
    @property
    def feeder_head_table(self):
        return self.output / TABLE_NAMES[0]
    
    @property
    def feeder_losses_table(self):
        return self.output / TABLE_NAMES[1]
    
    @property
    def metadata_table(self):
        return self.output / TABLE_NAMES[2]
    
    @property
    def thermal_metrics_table(self):
        return self.output / TABLE_NAMES[3]
    
    @property
    def voltage_metrics_table(self):
        return self.output / TABLE_NAMES[4]

    @property
    def snapshot_time_points_table(self):
        return self.output / TABLE_NAMES[5]

    @property
    def hosting_capacity_results(self):
        return self.output.glob("hosting_capacity_*.json")

    @property
    def pv_distances(self):
        return self.output / "weighted_average_pv_distances.csv"

    def __str__(self):
        """Return string representation of the output instance"""
        return str(self.output)

    @abstractmethod
    def validate_output(self, output):
        """Given an output, check if it contains desired reports
        
        Parameters
        ----------
        output: pathlib.Path

        Returns
        -------
        str, the valid output directory containing tables.
        """


class SimulationOutput(OutputBase):
    """Output instance class for handling Jade output"""
    
    @property
    def output_type(self):
        return OutputType.JADE
    
    def validate_output(self, output):
        """Check if output containing all desired reports
        
        Parameters
        ----------
        output: pathlib.Path
        """
        report_files = [output / table_name for table_name in self.table_names[:-1]]
        report_exists = [report_file.exists() for report_file in report_files]
        if not all(report_exists):
            raise ValueError(f"The output '{output}' does not contain valid reports.")
        return output


class PipelineSimulationOutput(OutputBase):
    """Output class for handling Jade pipeline output"""
    
    @property
    def output_type(self):
        return OutputType.JADE_PIPELINE
    
    def validate_output(self, output):
        """Return stage output that contains desired reports
        
        Parameters
        ----------
        output: str or pathlib.Path
        """
        for path in output.iterdir():
            if not path.name.startswith("output-stage"):
                continue
            report_files = [path / table_name for table_name in self.table_names[:-1]]
            report_exists = [report_file.exists() for report_file in report_files]
            if all(report_exists):
                return path
        
        raise IngestionError(f"No stage output in '{output}' contain valid reports")


class DermsSimulationOutput(OutputBase):

    @property
    def output_type(self):
        return OutputType.DERMS

    @property
    def result_file(self):
        for path in self.output.iterdir():
            if not path.is_dir():
                continue
            result_file = self.output / path / "results.json"
            if not result_file.exists():
                continue
            return result_file
        raise IngestionError("No JADE results.json file found")

    @property
    def derms_info_file(self):
        return self.output / DERMS_INFO_FILENAME

    def validate_output(self, output):
        report_files = [output / table_name for table_name in self.table_names[:-1]]
        report_exists = [report_file.exists() for report_file in report_files]
        if not all(report_exists):
            raise IngestionError(f"The output '{output}' does not contain valid reports.")
        return output


def get_creation_time(dir_or_file):
    if platform.system() == "Windows":
        return os.path.getctime(dir_or_file)
    
    stat = os.stat(dir_or_file)
    try:
        return stat.st_birthtime  # For FreeBSD, including Mac
    except AttributeError:
        # Not easy to get creation time on Linux, use modification time
        return stat.st_mtime


def is_from_pipeline(output):
    """Given an output, check if it's from pipeline or not
    
    Parameters
    ----------
    output: str or pathlib.Path
    """
    for path in output.iterdir():
        if path.name.startswith("output-stage"):
            return True
    return False


def is_from_derms(output):
    """Given an output, check if it's from DERMS simulation.
    
    Parameters
    ----------
    output: str or pathlib.Path
    """
    if not isinstance(output, pathlib.Path):
        output = pathlib.Path(output)
    
    derms_info_file = output / DERMS_INFO_FILENAME
    if derms_info_file.exists():
        return True
    return False


def get_simulation_output(output):
    """Given an output directory, return an instance of SimulationOutput or PipelineSimulationOutput
    
    Parameters
    ----------
    output: str or pathlib.Path
    """
    if not isinstance(output, pathlib.Path):
        output = pathlib.Path(output)
    
    if not output.exists():
        raise IngestionError(f"Output path does not exist - {output}.")
    
    if is_from_pipeline(output):
        output = PipelineSimulationOutput(output)
    elif is_from_derms(output):
        output = DermsSimulationOutput(output)
    else:
        output = SimulationOutput(output)
    
    logger.info("Simulation report tables are located in '%s'", output.output)
    
    return output



#!/usr/bin/env python

"""Maps elements to attached buses for each feeder in a path."""

import json
import logging
import os
import re
import sys

from jade.loggers import setup_logging
from jade.utils.utils import get_cli_string

REGION_BUS_MAPPING_FILENAME = "bus_mapping_summary.json"

logger = None


class MissingBus(Exception):
    pass


class BadFormat(Exception):
    pass


# This matches 1423_0_1409_0 and 242228 from
# New Capacitor.1423_0_1409_0 Bus1=242228 phases=3 Kv=4.8 conn=delta Kvar=450.0
REGEX_CAPACITOR = re.compile(r"New (?P<capacitor>Capacitor\.\w+).*Bus\w+=(?P<bus>\w+)", re.IGNORECASE)

# This matches 1418_0_7958_0, 242223, and 242246 from
# New Line.1418_0_7958_0 Units=km Length=0.15155920328177602 bus1=242223.1.2.3 bus2=242246.1.2.3 switch=n enabled=y phases=3 geometry=oh__314784__
REGEX_LINE = re.compile(r"New (?P<line>Line\.\w+).*bus\w+=(?P<bus_from>\w+).*bus\w+=(?P<bus_to>\w+)", re.IGNORECASE)

# This matches load_694781 and 242205_xfmr from
# New Load.load_694781 conn=delta bus1=242205_xfmr.1.2.3 kV=0.48 model=1 kW=3.225882583335277 kvar=0.7400682202361257 Phases=3
REGEX_LOAD = re.compile(r"New (?P<load>Load\.\w+).*bus\w+=(?P<bus>\w+)", re.IGNORECASE)

# This matches pv_865744 and 242208_xfmr from
# New PVSystem.pv_865744 bus1=242208_xfmr.1.2 phases=2 kV=0.20784609690826525 kVA=1.5829000000000002 Pmpp=1.439 conn=wye
REGEX_PV_SYSTEM = re.compile(r"New (?P<pv_system>PVSystem\.[\w-]+) bus\w+=(?P<bus>\w+)", re.IGNORECASE)

# This matches 1417_0_4346_0 and 3 from
# New Transformer.1417_0_4346_0 phases=1 windings=3 wdg=1 conn=delta bus=242214.2.3 Kv=4.8 kva=15.0 EmergHKVA=22.5 %r=0.1 wdg=2 conn=wye bus=242214_xfmr.1.0 Kv=0.12 kva=15.0 EmergHKVA=22.5 %r=0.1 wdg=3 conn=wye bus=242214_xfmr.0.2 Kv=0.12 kva=15.0 EmergHKVA=22.5 %r=0.1 XHL=0.1 XLT=0.1 XHT=0.1
REGEX_TRANSFORMER = re.compile(r"New (?P<transformer>Transformer\.\w+).*windings=(?P<windings>\d+).*bus=", re.IGNORECASE)
# After the above string has been split by "wdg=" this matches each wdg number and bus.
REGEX_TRANSFORMER_BUSES = re.compile(r"^(?P<winding>\d+).*conn=\w+.*bus=(?P<bus>\w+)", re.IGNORECASE)

# This matches an alternate transformer representation.
# New Transformer.trans_242188_reg phases=3 windings=2 buses=(242188_src.1.2.3,242188_dummy.1.2.3) conns=(Delta, Wye) kvs=(4.8, 4.8) kvas=(100000.0, 100000.0) XHL=0.1
REGEX_TRANSFORMER_ALT = re.compile(r"New (?P<transformer>Transformer\.\w+).*windings=(?P<windings>\d+).*buses=\((?P<buses>[\w\.,]+)\)", re.IGNORECASE)


def _check_bus(feeder_mapping, bus, name):
    if bus not in feeder_mapping["bus_coords"]:
        logger.warning("element=%s bus=%s is not present", name, bus)
        #raise MissingBus(f"bus={bus} element={elem_type}.{name} is not present")
        return False
    return True


def _check_transformer_buses(feeder_mapping, transformer, buses):
    if not _check_bus(feeder_mapping, buses["primary"], transformer):
        logger.warning("Missing coordinates for transformer=%s primary bus", transformer)
        return False

    if not _check_bus(feeder_mapping, buses["secondary"], transformer):
        logger.warning("Missing coordinates for transformer=%s secondary bus", transformer)
        feeder_mapping["bus_coords"][buses["secondary"]] = \
            feeder_mapping["bus_coords"][buses["primary"]]

    return True


def _check_bus_presence(feeder_mapping):
    for element_type, element_mapping in feeder_mapping.items():
        if element_type == "bus_coords":
            continue
        if element_type == "transformers":
            # Already checked.
            continue
        if element_type == "lines":
            for element, bus in element_mapping.items():
                _check_bus(feeder_mapping, bus["from"], element)
                _check_bus(feeder_mapping, bus["to"], element)
        else:
            for element, bus in element_mapping.items():
                _check_bus(feeder_mapping, bus, element)


def _handle_capacitor(feeder_mapping, line):
    match = REGEX_CAPACITOR.search(line)
    if not match:
        raise BadFormat(f"unsupported capacitor format: {line}")

    groupdict = match.groupdict()
    capacitor = groupdict["capacitor"]
    bus = groupdict["bus"]
    if capacitor in feeder_mapping["capacitors"]:
        logger.warning("Detected duplicate capacitor=%s", capacitor)
        return
    feeder_mapping["capacitors"][capacitor] = bus


def _handle_line(feeder_mapping, line):
    match = REGEX_LINE.search(line)
    if not match:
        raise BadFormat(f"unsupported line format: {line}")

    groupdict = match.groupdict()
    line_elem = groupdict["line"]
    bus_from = groupdict["bus_from"]
    bus_to = groupdict["bus_to"]
    if line_elem in feeder_mapping["lines"]:
        logger.warning("Detected duplicate line=%s", line_elem)
        return
    feeder_mapping["lines"][line_elem] = {
        "from": bus_from,
        "to": bus_to,
    }


def _handle_load(feeder_mapping, line):
    match = REGEX_LOAD.search(line)
    if not match:
        raise BadFormat(f"unsupported load format: {line}")

    groupdict = match.groupdict()
    load = groupdict["load"]
    bus = groupdict["bus"]
    if load in feeder_mapping["loads"]:
        logger.warning("Detected duplicate load=%s", load)
        return
    feeder_mapping["loads"][load] = bus


def _handle_pv_system(feeder_mapping, line):
    match = REGEX_PV_SYSTEM.search(line)
    if not match:
        raise BadFormat(f"unsupported pv_system format: {line}")

    groupdict = match.groupdict()
    pv_system = groupdict["pv_system"]
    bus = groupdict["bus"]
    if pv_system in feeder_mapping["pv_systems"]:
        assert bus == feeder_mapping["pv_systems"][pv_system]
    else:
        feeder_mapping["pv_systems"][pv_system] = bus


def _handle_transformer(feeder_mapping, line):
    match = REGEX_TRANSFORMER.search(line)
    if match:
        groupdict = match.groupdict()
        transformer = groupdict["transformer"]
        if transformer in feeder_mapping["transformers"]:
            logger.warning("Detected duplicate transformer=%s", transformer)
            return

        sections = line.split("wdg=")
        # First section does not contain wdg + bus.
        assert len(sections) == int(groupdict["windings"]) + 1, line
        buses = []
        for i, section in enumerate(sections):
            if i == 0:
                continue
            match = REGEX_TRANSFORMER_BUSES.search(section)
            assert match
            assert int(match.groupdict()["winding"]) == i
            bus = match.groupdict()["bus"]
            if bus not in buses:
                buses.append(bus)

        assert len(buses) >= 2
        t_buses = {
            "primary": buses[0],
            "secondary": buses[1],
        }
        _check_transformer_buses(feeder_mapping, transformer, t_buses)
        feeder_mapping["transformers"][transformer] = t_buses
    else:
        match = REGEX_TRANSFORMER_ALT.search(line)
        if not match:
            raise BadFormat(f"unsupported transformer format: {line}")

        groupdict = match.groupdict()
        transformer = groupdict["transformer"]
        if transformer in feeder_mapping["transformers"]:
            logger.warning("Detected duplicate transformer=%s", transformer)
            return

        buses = groupdict["buses"].split(",")
        assert len(buses) >= 2, line
        for i, bus in enumerate(buses):
            index = bus.find(".")
            if index == -1:
                raise BadFormat(f"unsupported bus format {bus}")
            buses[i] = bus[:index]
        t_buses = {
            "primary": buses[0],
            "secondary": buses[1],
        }
        _check_transformer_buses(feeder_mapping, transformer, t_buses)
        feeder_mapping["transformers"][transformer] = t_buses


def _make_feeder_dict():
    return {
        "bus_coords": {},
        "capacitors": {},
        "lines": {},
        "loads": {},
        "pv_systems": {},
        "transformers": {},
    }


def _process_bus_coords_file(mapping, filename):
    with open(filename) as f_in:
        for line in f_in:
            fields = line.split()
            if not fields:
                continue
            bus = fields[0]
            x = fields[1]
            y = fields[2]
            assert bus not in mapping
            mapping["bus_coords"][bus] = {"x": x, "y": y}


ELEMENT_HANDLERS = {
    "new capacitor.": _handle_capacitor,
    "new line.": _handle_line,
    "new load.": _handle_load,
    "new pvsystem.": _handle_pv_system,
    "new transformer.": _handle_transformer,
}


def _process_element_file(mapping, filename):
    with open(filename) as f_in:
        for line in f_in:
            for substring, handler in ELEMENT_HANDLERS.items():
                if substring in line.lower():
                    handler(mapping, line)


def _get_bus_coords_files(directory):
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            if filename.lower() == "buscoords.dss":
                yield os.path.join(dirpath, filename)


def _get_dss_files(directory):
    for dirpath, _, filenames in os.walk(directory):
        for filename in filenames:
            if os.path.splitext(filename)[1] == ".dss" and filename.lower() != "buscoords.dss":
                yield os.path.join(dirpath, filename)


def get_bus_to_element(bus_to_elems, feeder_mapping, element_type):
    """Builds a dict that maps bus names to a list of attached elements.

    Parameters
    ----------
    feeder_mapping : dict
        dictionary created by this module at a feeder level
    element_type : str
        Only consider this element type.
    bus_to_elems : dict
        Output dictionary to fill in.

    Returns
    -------
    dict
        Example::

        {123456: [{"type": "pv_systems", "name": "pv_123456"}]}

    """
    if element_type in ("transformers", "lines"):
        raise Exception("not supported")
    for element, bus in feeder_mapping[element_type].items():
        item = {"type": element_type, "name": element}
        if bus not in bus_to_elems:
            bus_to_elems[bus] = [item]
        else:
            bus_to_elems[bus].append(item)


def get_element_coordinates(feeder_mapping, element_type, name):
    """Return the coordinates of the bus to which the element is attached.

    Parameters
    ----------
    feeder_mapping : dict
        dictionary created by this module at a feeder level
    element_type : str
        capacitors, lines, loads, etc.
    name : str
        Element name

    Returns
    -------
    dict | None
        None is returned if no coordinates are stored.

        Example output::

        {'x': '34374.509', 'y': '206624.15'}

        If element_type == 'lines'
        {'from': None, 'to': {'x': '34802.251', 'y': '206769.654'}}

    """
    bus = feeder_mapping[element_type][name]
    if element_type == "lines":
        from_coords = feeder_mapping["bus_coords"].get(bus["from"])
        to_coords = feeder_mapping["bus_coords"].get(bus["to"])
        return {"from": from_coords, "to": to_coords}
    if element_type == "transformers":
        bus = bus["primary"]

    return feeder_mapping["bus_coords"].get(bus)


def make_element_bus_mapping_file(file_obj, output_file):
    """Make a mapping of elements to attached buses in a model-inputs path.

    Parameters
    ----------
    text : str
        Text of a .dss file

    Returns
    -------
    dict

    """
    feeder_mapping = _make_feeder_dict()
    _process_element_file(feeder_mapping, file_obj)

    # Much of the data has missing bus coordinates for transformer secondary
    # buses. The transformer processing code above implemented a workaround,
    # and so now they have coordinates.
    # Now, we check buses connected to the rest of the elements.
    _check_bus_presence(feeder_mapping)

    write_element_bus_mapping(feeder_mapping, output_file)
    print(f"Wrote {output_file}")
    return feeder_mapping


def merge_feeder_mappings(mapping_main, mapping_new):
    assert sorted(list(mapping_main.keys())) == sorted(list(mapping_new.keys()))
    for element_type in mapping_new.keys():
        if element_type == "bus_coords":
            continue
        for element, data in mapping_new[element_type].items():
            if element in mapping_main[element_type]:
                logger.info("Overwriting element %s %s", element_type, element)
            mapping_main[element_type][element] = data


def make_element_bus_mapping(path):
    """Make a mapping of elements to attached buses in a model-inputs path.

    Parameters
    ----------
    path : str
        Path to model-inputs directory (created by generate-input-data)

    Returns
    -------
    dict
        Maps feeder name to feeder mapping file.

    """
    config_file = os.path.join(path, "configurations.json")
    if not os.path.exists(config_file):
        logger.error("Did not detect %s in '%s'. Please pass a model-inputs directory.",
                     config_file, path)
        sys.exit(1)

    feeders = [x for x in os.listdir(path) if os.path.isdir(os.path.join(path, x))]

    feeder_files = {}

    # Get the bus coordinates first so that we can ensure that all element
    # buses are present.
    for feeder in feeders:
        feeder_mapping = _make_feeder_dict()
        feeder_path = os.path.join(feeder, path)
        for filename in _get_bus_coords_files(feeder_path):
            _process_bus_coords_file(feeder_mapping, filename)

        for filename in _get_dss_files(path):
            _process_element_file(feeder_mapping, filename)

        # Much of the data has missing bus coordinates for transformer
        # secondary buses. The transformer processing code above implemented a
        # workaround, and so now they have coordinates.  Now, we check buses
        # connected to the rest of the elements.
        _check_bus_presence(feeder_mapping)

        output_file = os.path.join(path, f"bus_mapping_feeder__{feeder}.json")
        write_element_bus_mapping(feeder_mapping, output_file)
        feeder_files[feeder] = output_file
        print(f"Wrote {output_file}")

    return feeder_files


def write_element_bus_mapping(mapping, output_file):
    """Write the mapping to output_file."""
    with open(output_file, "w") as f_out:
        json.dump(mapping, f_out)
        logger.info("Wrote %s", output_file)


def main():
    global logger
    if len(sys.argv) != 2:
        print(f"Usage:  {sys.argv[0]} MODEL-INPUTS-DIRECTORY")
        sys.exit(1)

    path = sys.argv[1]
    level = logging.INFO
    log_file = os.path.join(path, "bus_mapping.log")
    logger = setup_logging("bus_mapping", log_file, console_level=logging.ERROR, file_level=level,
                           packages=["disco"])
    logger.info(get_cli_string())

    try:
        feeder_files = make_element_bus_mapping(path)
        summary_file = os.path.join(path, REGION_BUS_MAPPING_FILENAME)
        with open(summary_file, "w") as f_out:
            json.dump(feeder_files, f_out, indent=2)
            print(f"Wrote summary file {summary_file}")
        print(f"Check {log_file} for warnings")
    except Exception:
        logger.exception("Failed to make bus mapping for %s", path)
        raise


if __name__ == "__main__":
    main()




import logging
from collections import namedtuple


from jade.exceptions import ExecutionError
from jade.result import ResultsSummary
from jade.utils.subprocess_manager import run_command
from jade.utils.utils import dump_data, load_data

from disco.common import EXIT_CODE_GOOD
from disco.exceptions import is_convergence_error
from disco.utils.failing_test_bisector import FailingTestBisector


JobKey = namedtuple("JobKeyByFeeder", ["substation", "feeder", "placement", "sample"])
PRESCREEN_JOBS_OUTPUT = "prescreen-jobs"


logger = logging.getLogger(__name__)


def create_job_key(job):
    deployment = job.model.deployment
    return JobKey(
        substation=deployment.substation,
        feeder=deployment.feeder,
        placement=deployment.project_data["placement"],
        sample=deployment.project_data["sample"],
    )


def run_prescreen(src_config, src_config_file, substation, feeder, placement, sample, prescreen_jobs_output):
    """Run a bisect on the penetration levels for a sample."""
    key = JobKey(substation, feeder, placement, sample)
    jobs = set()
    for job in src_config.iter_pydss_simulation_jobs(exclude_base_case=True):
        job_key = create_job_key(job)
        if key == job_key:
            jobs.add(job)

    jobs = list(jobs)
    jobs.sort(key=lambda x: x.model.deployment.project_data["penetration_level"])
    highest_level = find_highest_passing_penetration_level(key, jobs, src_config_file, prescreen_jobs_output)

    data = key._asdict()
    name = ("__".join((str(x) for x in data.values())))
    data["name"] = name
    data["highest_passing_penetration_level"] = highest_level
    filename = prescreen_jobs_output / (name + ".toml")
    dump_data(data, filename)
    logger.info("Created %s", filename)


def find_highest_passing_penetration_level(key, jobs, src_config_file, prescreen_jobs_output):
    job_names = []
    penetration_levels = []
    for job in jobs:
        job_names.append(job.name)
        penetration_levels.append(job.model.deployment.project_data["penetration_level"])

    bisector = FailingTestBisector(len(penetration_levels))
    index = bisector.get_first_index()
    done = False
    indices_run = set()
    for _ in range(len(penetration_levels)):
        name = job_names[index]
        run_config_file = create_job(src_config_file, name, prescreen_jobs_output)
        output_dir = prescreen_jobs_output / name
        ret = run_command(f"jade submit-jobs --local {run_config_file} -o {output_dir}")
        if ret != EXIT_CODE_GOOD:
            raise Exception(f"Unexpected JADE error occurred: key={key} name={name} ret={ret}")

        ret = _get_job_result(output_dir)
        if ret == EXIT_CODE_GOOD:
            passed = True
        elif is_convergence_error(ret):
            passed = False
        else:
            raise Exception(f"Unknown PyDSS error occurred: key={key} job={name}. End bisect: {ret}")

        indices_run.add(index)
        index, done = bisector.get_next_index(index, passed)
        if done:
            break
        if index in indices_run:
            raise Exception(f"already ran index={index} key={key} name={name}")

    if not done:
        raise Exception("Failed to find highest penetration level from {key}")

    return penetration_levels[index]


def create_job(config_file, name, prescreen_base_output):
    dst_config_file = prescreen_base_output / f"config__{name}.json"
    cmd = f"jade config filter --fields name {name} -o {dst_config_file} {config_file}"
    ret = run_command(cmd)
    if ret != 0:
        raise ExecutionError(f"filter failed: [{cmd}] ret={ret}")

    config = load_data(dst_config_file)
    assert len(config["jobs"]) == 1
    start_time = "2020-04-01T11:00:00.0"
    end_time = "2020-04-30T15:15:00.0"
    config["jobs"][0]["simulation"]["start_time"] = start_time
    config["jobs"][0]["simulation"]["end_time"] = end_time
    config["pydss_inputs"]["Simulation"]["project"]["simulation_range"] = {"start": "11:00:00", "end": "15:00:00"}
    config["pydss_inputs"]["Simulation"]["project"]["convergence_error_percent_threshold"] = 10.0

    config["pydss_inputs"]["Simulation"]["exports"]["export_results"] = False
    config["pydss_inputs"]["Simulation"]["exports"]["export_elements"] = False
    config["pydss_inputs"]["Simulation"]["reports"].clear()
    config["pydss_inputs"]["Scenarios"] = [
        x for x in config["pydss_inputs"]["Scenarios"] if x["name"] == "control_mode"
    ]
    assert config["pydss_inputs"]["Scenarios"]
    dump_data(config, dst_config_file)
    logger.info("Created prescreening job %s", dst_config_file)
    return dst_config_file


def _get_job_result(output_dir):
    result_summary = ResultsSummary(output_dir)
    results = result_summary.list_results()
    assert len(results) == 1
    result = results[0]
    return result.return_code



import json
import logging
import random
import sys
from types import SimpleNamespace

import click

from jade.loggers import setup_logging
from disco.enums import Placement
from disco.sources.base import DEFAULT_PV_DEPLOYMENTS_DIRNAME
from disco.sources.source_tree_1.pv_deployments import (
    DeploymentHierarchy,
    DeploymentCategory,
    PVDataStorage,
    PVDataManager,
    PVDeploymentManager,
    PVConfigManager
)

HIERARCHY_CHOICE = [item.value for item in DeploymentHierarchy]
CATEGORY_CHOICE = [item.value for item in DeploymentCategory]
PLACEMENT_CHOICE = [item.value for item in Placement]

logger = logging.getLogger(__name__)


def create_pv_deployments(input_path: str, hierarchy: str, config: dict):
    """A method for generating pv deployments"""
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    if not config.placement:
        print(f"'-p' or '--placement' should not be None for this action, choose from {PLACEMENT_CHOICE}")
        sys.exit()
    manager = PVDeploymentManager(input_path, hierarchy, config)
    summary = manager.generate_pv_deployments()
    print(json.dumps(summary, indent=2))


def create_pv_configs(input_path: str, hierarchy: str, config: dict, control_name: str, limit: int):
    """A method for generating pv config JSON files """
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    if config.placement:
        config.placement = None
        print(f"'-p' or '--placement' option is ignored for this action.")
    
    manager = PVConfigManager(input_path, hierarchy, config)
    config_files = manager.generate_pv_configs(control_name=control_name, limit=limit)
    print(f"PV configs created! Total: {len(config_files)}")


def remove_pv_deployments(input_path: str, hierarchy: str, config: dict):
    """A method for removing deployed pv systems"""
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVDeploymentManager(input_path, hierarchy, config)
    if config.placement:
        placement = Placement(config.placement)
    else:
        placement = config.placement
    result = manager.remove_pv_deployments(placement=placement)
    print(f"=========\nTotal removed deployments: {len(result)}")


def check_pv_deployments(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVDeploymentManager(input_path, hierarchy, config)
    if config.placement:
        placement = Placement(config.placement)
    else:
        placement = config.placement
    result = manager.check_pv_deployments(placement=placement)
    print(json.dumps(result.__dict__, indent=2))


def remove_pv_configs(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVConfigManager(input_path, hierarchy, config)
    if config.placement:
        placement = Placement(config.placement)
    else:
        placement = config.placement
    config_files = manager.remove_pv_configs(placement=placement)
    print(f"PV configs created! Total: {len(config_files)}")


def check_pv_configs(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVConfigManager(input_path, hierarchy, config)
    if config.placement:
        placement = Placement(config.placement)
    else:
        placement = config.placement
    total_missing = manager.check_pv_configs(placement=placement)
    print(json.dumps(total_missing, indent=2))


def list_feeder_paths(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    storage = PVDataStorage(input_path, hierarchy, config)
    result = storage.get_feeder_paths()
    for feeder_path in result:
        print(feeder_path)
    print(f"=========\nTotal feeders: {len(result)}")


def redirect_pv_shapes(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVDataManager(input_path, hierarchy, config)
    if hierarchy == DeploymentHierarchy.SUBSTATION:
        manager.redirect_substation_pv_shapes()
    elif hierarchy == DeploymentHierarchy.FEEDER:
        manager.redirect_feeder_pv_shapes()
    else:
        raise NotImplementedError(f"{hierarchy=}")


def generate_pv_deployment_jobs(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    
    manager = PVDeploymentManager(input_path, hierarchy, config)
    manager.generate_pv_creation_jobs()
    
    manager = PVConfigManager(input_path, hierarchy, config)
    manager.generate_pv_config_jobs()


def restore_feeder_data(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVDataManager(input_path, hierarchy, config)
    manager.restore_feeder_data()


def transform_feeder_loads(input_path: str, hierarchy: str, config: dict):
    hierarchy = DeploymentHierarchy(hierarchy)
    config = SimpleNamespace(**config)
    manager = PVDataManager(input_path, hierarchy, config)
    manager.transform_feeder_loads()


ACTION_MAPPING = {
    "redirect-pvshapes": redirect_pv_shapes,
    "transform-loads": transform_feeder_loads,
    "generate-jobs": generate_pv_deployment_jobs,
    "restore-feeders": restore_feeder_data,
    
    "create-pv": create_pv_deployments,
    "create-configs": create_pv_configs,
    
    "remove-pv": remove_pv_deployments,
    "check-pv": check_pv_deployments,
    "remove-configs": remove_pv_configs,
    "check-configs": check_pv_configs,
    "list-feeders": list_feeder_paths
}

@click.group()
def pv_deployments():
    """Generate PV deployments from raw OpenDSS models"""


@click.command()
@click.argument("input_path")
@click.option(
    "-a", "--action",
    type=click.Choice(list(ACTION_MAPPING.keys()), case_sensitive=False),
    required=True,
    help="Choose the action related to pv deployments"
)
@click.option(
    "-c", "--control-name",
    type=click.STRING,
    default="volt_var_ieee_1547_2018_catB",
    show_default=True,
    help="Choose the control name to assign to pv configs in the action create-configs."
)
@click.option(
    "-h", "--hierarchy",
    type=click.Choice(HIERARCHY_CHOICE, case_sensitive=False),
    required=True,
    help="Choose the deployment hierarchy."
)
@click.option(
    "-l", "--kw-limit",
    type=click.FLOAT,
    default=5,
    show_default=True,
    help="Capacity threshold to use for assigning the value of --control-name. The action "
    "create-configs will only assign a control if a PV's capacity is greater than this value in "
    "kW.",
)
@click.option(
    "-p", "--placement",
    type=click.Choice(PLACEMENT_CHOICE, case_sensitive=False),
    required=False,
    default=None,
    show_default=True,
    help="Choose the placement type"
)
@click.option(
    "-c", "--category",
    type=click.Choice(CATEGORY_CHOICE, case_sensitive=False),
    default=DeploymentCategory.SMALL.value,
    show_default=True,
    help="The PV size pdf value"
)
@click.option(
    "-f", "--master-filename",
    type=click.STRING,
    required=False,
    default="Master.dss",
    show_default=True,
    help="The filename of master dss"
)
@click.option(
  "-m", "--min-penetration",
    type=click.INT,
    default=5,
    show_default=True,
    help="The minimum level of PV penetration."
)
@click.option(
    "-M", "--max-penetration",
    type=click.INT,
    default=200,
    show_default=True,
    help="The maximum level of PV penetration."
)
@click.option(
   "-s", "--penetration-step",
    type=click.INT,
    default=5,
    show_default=5,
    help="The step of penetration level."
)
@click.option(
    "-n", "--sample-number",
    type=click.INT,
    default=10,
    show_default=True,
    help="The number of deployments"
)
@click.option(
    "-S", "--proximity-step",
    type=click.INT,
    default=10,
    show_default=True,
    help="The proximity step in PV deployments."
)
# TODO: Jianli, this code option causes an exception with the command below.
# The option is not used, so I'm commenting it out.
#@click.option(
#    "-P", "--percent-shares",
#    default=[100, 0],
#    show_default=True,
#    help="The share pair - [share of residential PVs, share of utility scale PVs]"
#)
@click.option(
    "-x", "--pv-size-pdf",
    type=click.INT,
    default=None,
    show_default=True,
    help="The PV size pdf value"
)
@click.option(
    "--pv-upscale/--no-pv-upscale",
    is_flag=True,
    default=True,
    show_default=True,
    help="Upscale PV in deployments."
)
@click.option(
    "-o", "--pv-deployments-dirname",
    type=click.STRING,
    default=DEFAULT_PV_DEPLOYMENTS_DIRNAME,
    show_default=True,
    help="Output directory name of PV deployments"
)
@click.option(
    "-r", "--random-seed",
    type=click.IntRange(1, 1000000, clamp=True),
    default=random.randint(1, 1000000),
    help="Set an initial integer seed for making PV deployments reproducible"
)
@click.option(
    "--verbose",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable to show overbose information."
)
def source_tree_1(
    input_path,
    action,
    control_name,
    hierarchy,
    kw_limit,
    placement,
    category,
    master_filename,
    min_penetration,
    max_penetration,
    penetration_step,
    sample_number,
    proximity_step,
    #percent_shares,
    pv_size_pdf,
    pv_upscale,
    pv_deployments_dirname,
    random_seed,
    verbose
):
    """Generate PV deployments for source tree 1."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging("pv_deployments", None, console_level=level, packages=["disco"])
    
    if action == "create-pv":
        logger.info("Set integer %s as initial random seed for PV deployments.", random_seed)
    
    if action == "create-configs":
        logger.info("Set integer %s as initial random seed for PV configs.", random_seed)
    
    config = {
        "placement": placement,
        "category": category,
        "master_filename": master_filename,
        "pv_upscale": pv_upscale,
        "min_penetration": min_penetration,
        "max_penetration": max_penetration,
        "penetration_step": penetration_step,
        "sample_number": sample_number,
        "proximity_step": proximity_step,
        "percent_shares": [100, 0],
        "pv_size_pdf": pv_size_pdf,
        "pv_deployments_dirname": pv_deployments_dirname,
        "random_seed": random_seed
    }
    action_function = ACTION_MAPPING[action]
    args = [input_path, hierarchy, config]
    if action == "create-configs":
        args.append(control_name)
        args.append(kw_limit)
    action_function(*args)


pv_deployments.add_command(source_tree_1)




import copy
import logging

from disco.extensions.pydss_simulation.pydss_configuration import \
    PyDssConfiguration
from disco.extensions.pydss_simulation.pydss_inputs import PyDssInputs
from disco.pydss.common import ConfigType
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


def auto_config(inputs, **kwargs):
    """Create a configuration from all available inputs."""
    if isinstance(inputs, str):
        inputs = PyDssInputs(inputs)
    config = PyDssConfiguration(inputs, **kwargs)
    for job in config.inputs.iter_jobs():
        config.add_job(job)

    #exports = load_data(exports_filename)
    #config.set_pydss_config(ConfigType.EXPORTS, exports)

    return config



"""Enums for the disco package."""

import enum
import re

import toml


class Status(enum.Enum):
    """Return status."""
    GOOD = 0
    ERROR = 1


# Directories use these values.
DCAC_MAPPING = {
    -1: "None",
    0.8: "DCAC0.8",
    1.15: "DCAC1.15",
}


class Scale(enum.Enum):
    """Possible values for scale"""
    SMALL = "small"
    LARGE = "large"


# Directories use these values.
SCALE_MAPPING = {
    Scale.SMALL: "SmallScale",
    Scale.LARGE: "LargeScale",
}


def get_scale_from_value(value):
    """Gets the enum for the given value.

    Returns
    -------
    Scale

    """
    for key, val in SCALE_MAPPING.items():
        if val == value:
            return key
    raise Exception("Unknown value: {}".format(value))


class Placement(enum.Enum):
    """Possible values for placement"""
    RANDOM = "random"
    CLOSE = "close"
    FAR = "far"


# TODO: we cannot support having placement values with upper and lower case.
# need to change all directories where this is the case.


def get_placement_from_value(value):
    """Gets the enum for the given value.

    Note: this performs a case-insensitive search because some input
    directories are inconsistent.

    Returns
    -------
    Placement

    """
    for placement in Placement:
        if placement.value == value.lower():
            return placement
    raise Exception("Unknown value: {}".format(value))


class Mode(enum.Enum):
    """Possible values for computational sequencing mode"""
    PARALLEL = "parallel"
    SEQUENTIAL = "sequential"


class SimulationHierarchy(enum.Enum):
    SUBSTATION = "substation"
    FEEDER = "feeder"


class SimulationOutputs(enum.Enum):
    """Defines the possible simulation output types."""
    LINE_LOADINGS = "line_loadings"
    LINE_LOSSES = "line_losses"
    NODAL_VOLTAGES = "nodal_voltages"
    PV_OUTPUTS = "pv_outputs"
    TOTAL_CIRCUIT_POWER = "total_circuit_power"
    TOTAL_LOSSES = "total_losses"
    TRANSFORMER_LOADINGS = "transformer_loadings"


class SimulationType(enum.Enum):
    """Defines the possible simulation types."""
    SNAPSHOT = "snapshot"
    QSTS = "QSTS"
    TIME_SERIES = "time-series"  # This should be duplicate with QSTS.
    UPGRADE = "upgrade"


class AnalysisModelType(enum.Enum):
    """Represents the type of model used for each analysis type."""
    # ImpactAnalysis is improperly used here. Leaving for backwards compatibility.
    # The same data model is used for time-series impact analysis and hosting capacity analysis.
    SnapshotImpactAnalysis = "SnapshotImpactAnalysis"
    TimeSeriesImpactAnalysis = "TimeSeriesImpactAnalysis"
    UpgradeCostAnalysis = "UpgradeCostAnalysis"


class AnalysisType(enum.Enum):
    """DISCO user analysis types"""
    IMPACT_ANALYSIS = "impact-analysis"
    HOSTING_CAPACITY = "hosting-capacity"
    COST_BENEFIT = "cost-benefit"
    UPGRADE_ANALYSIS = "upgrade-analysis"
    NONE = "none"


class LoadMultiplierType(enum.Enum):
    """Multiplier types"""
    ORIGINAL = "original"
    UNIFORM = "uniform"


ANALYSIS_MODEL_TYPES = [t.value for t in AnalysisModelType]


PUBLIC_ENUMS = {
    "AnalysisModelType": AnalysisModelType,
    "Mode": Mode,
    "Placement": Placement,
    "Scale": Scale,
    "SimulationOutputs": SimulationOutputs,
    "SimulationType": SimulationType,
}


_REGEX_ENUM = re.compile(r"^(\w+)\.(\w+)$")


def get_enum_from_str(string):
    """Converts a enum that's been written as a string back to an enum.

    Parameters
    ----------
    string : str
        string to convert

    Returns
    -------
    tuple
        converted, converted enum or original

    """
    converted = False
    match = _REGEX_ENUM.search(string)
    if match and match.group(1) in PUBLIC_ENUMS:
        obj = getattr(PUBLIC_ENUMS[match.group(1)], match.group(2))
        converted = True
    else:
        obj = string

    return converted, obj


def get_enum_from_value(cls, value):
    """Gets the enum for the given value."""
    for enum_ in cls:
        if enum_.value == value:
            return enum_
    raise Exception("Unknown value: {} {}".format(cls, value))


class EnumEncoder(toml.TomlEncoder):
    """Custom encoder for enums."""
    def __init__(self, _dict=dict, preserve=False):
        super(EnumEncoder, self).__init__(_dict, preserve)
        self.dump_funcs[enum.Enum] = EnumEncoder.dump_enum

    @staticmethod
    def dump_enum(val):
        """Return the enum value converted to string."""
        return str(val.value)



"""
setup.py
"""
import os
import logging
from codecs import open
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
import sys


logger = logging.getLogger(__name__)


def read_lines(filename):
    return Path(filename).read_text().splitlines()


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        install_jade_extensions()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        install_jade_extensions()


def install_jade_extensions():
    # These won't be available until after jade gets installed by pip.
    from jade.extensions.registry import Registry
    from jade.utils.subprocess_manager import run_command

    registry_filename = Path.home() / Registry._REGISTRY_FILENAME
    if os.path.exists(registry_filename):
        os.remove(registry_filename)
    ext = os.path.join(here, "disco", "extensions", "jade_extensions.json")
    run_command(f"jade extensions register {ext}")
    run_command("jade extensions add-logger disco")


here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "disco", "version.py"), encoding="utf-8") as f:
    lines = f.read().split("\n")
    if len(lines) != 2:
        print("Invalid format in version.py", file=sys.stderr)
        sys.exit(1)


version = lines[0].split()[2].strip('"').strip("'")

install_requires = [
    "NREL-jade~=0.10.0",
    "chevron~=0.14.0",
    "click~=8.0",
    "dsspy==3.0.2",  # 3.0.4 has a breaking change
    "filelock~=3.8",
    "matplotlib~=3.6",
    "networkx~=2.8",
    "opendssdirect.py~=0.8.4",
    "openpyxl~=3.0",
    "pandas >= 2, < 3",
    "pydantic~=2.5.2",
    "seaborn~=0.12.1",
    "scikit-learn >= 1.1, < 1.4",  # 1.4 removes the parameter affinity from AgglomerativeClustering
    "sqlalchemy~=1.4",
    "toml~=0.10.0",
]
dev_requires = [
    "flake8",
    "ghp-import",
    "mock>=3.0.0",
    "pycodestyle",
    "pylint",
    "pytest",
    "pytest-cov",
    "sphinx>=2.0",
    "sphinx-rtd-theme>=0.4.3",
    "sphinxcontrib-plantuml",
    "tox",
]
test_requires = ["pytest"]

setup(
    name="NREL-disco",
    version=version,
    description="DISCO",
    long_description=readme,
    long_description_content_type="text/markdown",
    maintainer_email="daniel.thom@nrel.gov",
    url="https://github.com/NREL/disco",
    packages=find_packages(),
    package_dir={"disco": "disco"},
    entry_points={
        "console_scripts": [
            "disco=disco.cli.disco:cli",
            "disco-internal=disco.cli.disco_internal:cli",
        ],
    },
    include_package_data=True,
    package_data={
        "disco": [
            "analysis/*.toml",
            "analysis/*.xlsx",
            "pipelines/template/*.toml",
            "extensions/pydss_simulation/*.toml",
            "extensions/pydss_simulation/trained_lm_time_prediction.sav",
            "extensions/upgrade_simulation/upgrades/*.xlsx",
            "extensions/upgrade_simulation/upgrades/*.toml",
            "extensions/*.json",
            "postprocess/config/*.toml",
            "postprocess/toolbox/query_tool.ipynb",
            "pydss/config/*.toml",
        ],
    },
    license="BSD license",
    zip_safe=False,
    keywords=["disco"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
        "extras": ["ipywidgets"]
    },
    # Disabled because this method is not compatible with wheels, and so we
    # can't build a PyPi package.
    #cmdclass={"install": PostInstallCommand, "develop": PostDevelopCommand},
)



#!/usr/bin/env python

"""Creates JADE configuration for stage 1 of pydss_simulation pipeline."""

import logging
import sys

import click

from jade.common import CONFIG_FILE
from jade.loggers import setup_logging
from jade.utils.utils import load_data
from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO

from disco.enums import SimulationType
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.extensions.pydss_simulation.estimate_run_minutes import generate_estimate_run_minutes
from disco.pydss.common import ConfigType
from disco.pydss.pydss_configuration_base import get_default_reports_file

logger = logging.getLogger(__name__)


def callback_is_enabled(_, __, value):
    if value is None:
        return None
    return {"true": True, "false": False}[value.lower()]


COMMON_TIME_SERIES_OPTIONS = (
    click.option(
        "-c",
        "--config-file",
        default=CONFIG_FILE,
        show_default=True,
        help="JADE config file to create",
    ),
    click.option(
        "--feeder-losses",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the Feeder Losses report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--pv-clipping",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the PV clipping report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--pv-curtailment",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the PV curtailment report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--thermal-metrics",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the Thermal Metrics report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--voltage-metrics",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the Voltage Metrics report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--capacitor-changes",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the Capacitor State Changes report. If not set, use the value in "
        "--reports-filename.",
    ),
    click.option(
        "--regcontrol-changes",
        type=click.Choice(("true", "false"), case_sensitive=False),
        callback=callback_is_enabled,
        default=None,
        show_default=True,
        help="Whether to enable the RegControl Tap Number Changes report. If not set, use the "
        "value in --reports-filename.",
    ),
    click.option(
        "--export-data-tables",
        default=False,
        is_flag=True,
        show_default=True,
        help="Export collected circuit element properties as tables.",
    ),
    click.option(
        "--exports-filename",
        default=None,
        show_default=True,
        help="PyDSS export options, default is None.",
    ),
    click.option(
        "-r",
        "--reports-filename",
        default=get_default_reports_file(SimulationType.QSTS),
        show_default=True,
        help="PyDSS report options",
    ),
    click.option(
        "--skip-night/--no-skip-night",
        default=False,
        is_flag=True,
        show_default=True,
        help="Don't run controls or collect data during nighttime hours.",
    ),
    click.option(
        "--store-all-time-points/--no-store-all-time-points",
        is_flag=True,
        default=False,
        show_default=True,
        help="Store per-element data at all time points for thermal and voltage metrics.",
    ),
    click.option(
        "--store-per-element-data/--no-store-per-element-data",
        is_flag=True,
        default=False,
        show_default=True,
        help="Store per-element data in thermal and voltage metrics.",
    ),
    click.option(
        "-v",
        "--volt-var-curve",
        default=None,
        help="Update the PyDSS volt-var curve name. If not set, use the pre-configured curve.",
    ),
    click.option(
        "--verbose",
        is_flag=True,
        default=False,
        help="Enable debug logging",
    ),
)


def common_time_series_options(func):
    for option in reversed(COMMON_TIME_SERIES_OPTIONS):
        func = option(func)
    return func


@click.command()
@click.argument("inputs")
@common_time_series_options
@click.option(
    "-e",
    "--estimated-run-minutes",
    type=int,
    help="Estimated per-job runtime. Default is None.",
)
@click.option(
    "--calc-estimated-run-minutes/--no-calc-estimated-run-minutes",
    is_flag=True,
    default=True,
    show_default=True,
    help="Calculate estimated per-job runtime by parsing the OpenDSS files.",
)
@click.option(
    "--dc-ac-ratio",
    default=None,
    type=float,
    help="Set a custom DC-AC ratio for PV Systems.",
)
@click.option(
    "--pf1/--no-pf1",
    is_flag=True,
    default=True,
    show_default=True,
    help="Include PF1 scenario or not",
)
@click.option(
    "--control-mode/--no-control-mode",
    is_flag=True,
    default=True,
    show_default=True,
    help="Include control_mode scenario or not",
)
@click.option(
    "--order-by-penetration/--no-order-by-penetration",
    default=False,
    show_default=True,
    help="Make jobs with higher penetration levels blocked by those with lower levels. This "
    "can be beneficial if you want the higher-penetration-level jobs to be "
    "canceled if a job with a lower penetration level fails. However, it can significantly "
    "reduce the number of jobs that can run simultaneously.",
)
def time_series(
    inputs,
    config_file,
    feeder_losses,
    pv_clipping,
    pv_curtailment,
    thermal_metrics,
    voltage_metrics,
    capacitor_changes,
    regcontrol_changes,
    export_data_tables,
    exports_filename,
    reports_filename,
    skip_night,
    store_all_time_points,
    store_per_element_data,
    volt_var_curve,
    verbose,
    estimated_run_minutes,
    calc_estimated_run_minutes,
    dc_ac_ratio,
    pf1,
    control_mode,
    order_by_penetration,
):
    """Create JADE configuration for time series simulations."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])

    if not pf1 and not control_mode:
        logger.error("At least one of '--pf1' or '--control-mode' must be set.")
        sys.exit(1)

    simulation_config = PyDssConfiguration.get_default_pydss_simulation_config()
    simulation_config["project"]["simulation_type"] = SimulationType.QSTS.value
    simulation_config["reports"] = load_data(reports_filename)["reports"]
    simulation_config["exports"]["export_data_tables"] = export_data_tables
    for report in simulation_config["reports"]["types"]:
        if report["name"] == "Feeder Losses" and feeder_losses is not None:
            report["enabled"] = feeder_losses
        if report["name"] == "PV Clipping" and pv_clipping is not None:
            report["enabled"] = pv_clipping
        if report["name"] == "PV Curtailment" and pv_curtailment is not None:
            report["enabled"] = pv_curtailment
        if report["name"] == "Thermal Metrics" and thermal_metrics is not None:
            report["enabled"] = thermal_metrics
        if report["name"] == "Voltage Metrics" and voltage_metrics is not None:
            report["enabled"] = voltage_metrics
        if report["name"] in ("Thermal Metrics", "Voltage Metrics"):
            report["store_all_time_points"] = store_all_time_points
            report["store_per_element_data"] = store_per_element_data
        if report["name"] == "Capacitor State Change Counts" and capacitor_changes is not None:
            report["enabled"] = capacitor_changes
        if report["name"] == "RegControl Tap Number Change Counts" and regcontrol_changes is not None:
            report["enabled"] = regcontrol_changes

    exports = {} if exports_filename is None else load_data(exports_filename)
    scenarios = []
    if control_mode:
        scenarios.append(
            PyDssConfiguration.make_default_pydss_scenario(CONTROL_MODE_SCENARIO, exports)
        )
    if pf1:
        scenarios.append(PyDssConfiguration.make_default_pydss_scenario(PF1_SCENARIO, exports))
    config = PyDssConfiguration.auto_config(
        inputs,
        simulation_config=simulation_config,
        scenarios=scenarios,
        order_by_penetration=order_by_penetration,
        estimated_run_minutes=estimated_run_minutes,
        dc_ac_ratio=dc_ac_ratio,
    )

    has_pydss_controllers = config.has_pydss_controllers()
    if control_mode and not has_pydss_controllers:
        scenarios_config = config.get_pydss_config(ConfigType.SCENARIOS)
        assert scenarios_config[0]["name"] == CONTROL_MODE_SCENARIO
        scenarios_config.pop(0)
        logger.info(
            "Excluding %s scenario because there are no pydss controllers.", CONTROL_MODE_SCENARIO
        )
        config.set_pydss_config(ConfigType.SCENARIOS, scenarios_config)

    if volt_var_curve is not None:
        if has_pydss_controllers and control_mode:
            config.update_volt_var_curve(volt_var_curve)
        else:
            logger.warning(
                "Setting a volt_var_curve has no effect when there is no %s scenario.",
                CONTROL_MODE_SCENARIO,
            )

    if calc_estimated_run_minutes:
        generate_estimate_run_minutes(config)

    if skip_night:
        pydss_sim_config = config.get_pydss_config(ConfigType.SIMULATION_CONFIG)
        pydss_sim_config["project"]["simulation_range"] = {"start": "06:00:00", "end": "18:00:00"}
        # Note that we are using the same convergence error threshold percent.
        config.set_pydss_config(ConfigType.SIMULATION_CONFIG, pydss_sim_config)

    config.dump(filename=config_file)

    print(f"Created {config_file} for TimeSeries Analysis")



"""Defines Analysis objects."""

from abc import ABC, abstractmethod
from numpy import bool_

from jade.utils.utils import load_data


class Analysis(ABC):
    """Base Analysis class containing required inputs and options

    Attributes
    ----------
    inputs : list (default empty list)
    options : list (default empty list)

    """
    INPUTS = []
    OPTIONS = []

    def __init__(self, overrides=None, job_name=None):
        if overrides is not None:
            self.set_config_overrides(overrides)
        self._job_name = job_name
        self._results = []

    @abstractmethod
    def run(self):
        """Function to run when analysis has been built"""

    def get_option(self, option_name):
        """Return option with given name
        Parameters
        ----------
        option_name : str

        Returns
        -------
        Option

        """
        for option_config in self.get_options():
            if option_config.name == option_name:
                return option_config

        return None

    def get_input(self, input_name):
        """Return input with given name
        Parameters
        ----------
        input_name : str

        Returns
        -------
        Input

        """
        for input_config in self.get_inputs():
            if input_config.name == input_name:
                return input_config

        return None

    def print_inputs(self):
        """Prints name, type and default values for all inputs"""
        for analysis_input in self.INPUTS:
            name = analysis_input.name
            input_type = analysis_input.input_type.get_type_name()
            default_value = analysis_input.default

            if input_type == "dict":
                maxlen = max([len(k) for k in default_value.keys()])
                template = "\n{:<{width}}   {}\n".format("Parameter", "Value", width=maxlen)
                for k, v in default_value.items():
                    if v is True:
                        v = "true"
                    elif v is False:
                        v = "false"
                    template += "{:<{width}} : {}\n".format(k, str(v), width=maxlen)
                default_value = template.rstrip()

            print(f"Name: {name}\n" +
                  f"Type: ({input_type})\n" +
                  f"Default: {default_value}\n")

    def get_options(self):
        """Return option list"""
        return self.OPTIONS[:]

    def get_inputs(self):
        """Return input list"""
        return self.INPUTS[:]

    def serialize_inputs(self):
        """Return serialized input list"""
        return [input.serialize() for input in self.INPUTS]

    def set_option_overrides(self, overrides):
        """Set option values, given optional list of overrides

        Parameters
        ----------
        overrides : list
            list of overrides

        """
        for index, required_option in enumerate(self.OPTIONS):
            user_option = None
            if required_option.name in overrides.keys() :
                user_option = overrides[required_option.name]

            if user_option is None or user_option == '':
                parsed_value = required_option.default

            required_option.value = parsed_value
            self.OPTIONS[index] = required_option

    def set_config_overrides(self, overrides):
        """Set config values, given dict of overrides

        Parameters
        ----------
        overrides : dict

        Raises
        ------
        ValueError
            Raised if given input does not match type of current content

        """
        for index, required_input in enumerate(self.INPUTS):
            user_input = overrides.get(required_input.name)
            if user_input is None or user_input == '':
                parsed_value = required_input.default
            else:
                parsed_value = required_input.input_type.parse_value(user_input)

            if parsed_value is None:
                raise ValueError('Please input valid value for ' + required_input.name)

            required_input.value = parsed_value
            self.INPUTS[index] = required_input

    def get_results(self):
        """Returns analysis results"""
        inputs = self.serialize_inputs()
        # remove default index
        for i in inputs:
            i.pop('default')
        return {'inputs': inputs, 'outputs': self._results}

    @property
    def serialized_data(self):
        """Returns all INPUTS, user or defaults, as a dictionary"""
        serialized_data = {}
        for data in self.serialize_inputs():
            serialized_data[data['name']] = data['value']
        return serialized_data

    def _add_to_results(self, result_type, data):
        """Appends given results to list of result dicts

        Parameters
        ----------
        result_type: str
        data: dict

        """
        new_result = {'result_type': result_type, 'data': serialize_results(data)}
        self._results.append(new_result)


def load_config_file(_, __, value):
    """Load config file
    Parameters
    ----------
    value : str
        config file location

    Returns
    -------
    dict
        loaded config file

    """
    if value is None:
        return None

    return load_data(value)


def load_custom_overrides(_, __, value):
    """Load custom overrides from user
    Parameters
    ----------
    value : list
        list of str (override_name=override_value)

    Returns
    -------
    dict
        user overrides

    """
    user_overrides = {}

    for override in value:
        override_name, override_value = override.split('=')
        user_overrides[override_name] = override_value
    return user_overrides


def serialize_results(data):
    """Serialize results to JSON parse-able values

    Parameters
    ----------
    data : dict

    Returns
    -------
    dict

    """
    if not isinstance(data, dict) and not isinstance(data, list):
        return data

    serialized = {}
    for key, value in data.items():
        if isinstance(value, bool_):
            value = bool(value)
        serialized[key] = value
    return serialized



"""Command to download input data."""

import click

from disco.sources.epri import (
    download_epri_feeder_opendss_data,
    define_epri_source_format
)


@click.group()
def download_source():
    """Commands to download raw OpenDSS models from the source."""


@click.command()
@click.argument("feeders", nargs=-1)
@click.option(
    "-d", "--directory",
    default=".",
    show_default=True,
    help="local download directory",
)
def epri(feeders, directory):
    """Download EPRI feeder models - J1, K1, and/or M1"""
    for feeder in feeders:
        location = download_epri_feeder_opendss_data(feeder, output_directory=directory)
        print(f"Downloaded EPRI {feeder} to {location}")
    
    define_epri_source_format(output_directory=directory)


download_source.add_command(epri)



"""Model for Source Type 1 feeder input data"""

import copy
import fileinput
import itertools
import json
import logging
import os
import re
from collections import namedtuple, defaultdict
from pathlib import Path

import click

from jade.exceptions import InvalidParameter
from jade.utils.utils import ExtendedJSONEncoder
from disco.cli.common import handle_existing_dir
from disco.enums import Placement, SimulationType, SimulationHierarchy
from disco.models.base import PyDSSControllerModel, ImpactAnalysisBaseModel
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.models.time_series_analysis_model import TimeSeriesAnalysisModel
from disco.models.upgrade_cost_analysis_model import UpgradeCostAnalysisModel
from disco.sources.base import (
    BaseOpenDssModel,
    SOURCE_CONFIGURATION_FILENAME,
    DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS,
    DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS,
    DEFAULT_PV_DEPLOYMENTS_DIRNAME,
    DEFAULT_UPGRADE_COST_ANALYSIS_PARAMS
)
from .source_tree_1_model_inputs import SourceTree1ModelInputs


logger = logging.getLogger(__name__)


SubstationKey = namedtuple("SampleKey", "substation, placement, sample, penetration_level")


def _process_hierarchy(_, __, val):
    return SimulationHierarchy(val)


COMMON_OPTIONS = (
    click.option(
        "--hierarchy",
        type=click.Choice([x.value for x in SimulationHierarchy]),
        callback=_process_hierarchy,
        default=SimulationHierarchy.FEEDER.value,
        show_default=True,
        help="Level at which to configure the simulation",
    ),
    click.option(
        "-S",
        "--substations",
        default=["all"],
        multiple=True,
        show_default=True,
        help="substations to add; use ('all',) to auto-detect and add all substations",
    ),
    click.option(
        "-f",
        "--feeders",
        default=["all"],
        multiple=True,
        show_default=True,
        help="feeders to add; use ('all',) to auto-detect and add all feeders",
    ),
    click.option(
        "-p",
        "--placements",
        default=["all"],
        multiple=True,
        show_default=True,
        help="placements to add; use ('all',) to auto-detect and add all placements",
    ),
    click.option(
        "-d",
        "--samples",
        default=["all"],
        multiple=True,
        show_default=True,
        help="samples to add; use ('all',) to auto-detect and add all samples",
    ),
    click.option(
        "-l",
        "--penetration-levels",
        default=["all"],
        multiple=True,
        show_default=True,
        help="penetration-levels to add; use ('all',) to auto-detect and add all penetration-levels",
    ),
    click.option(
        "-m",
        "--master-file",
        default="Master.dss",
        show_default=True,
        help="Master file for the OpenDSS model",
    ),
    click.option(
        "-c",
        "--copy-load-shape-data-files",
        help="Copy load shape data files into the target directory. Otherwise, use references "
             "to existing files.",
        is_flag=True,
        default=False,
        show_default=True,
    ),
    click.option(
        "-x",
        "--strip-load-shape-profiles",
        is_flag=True,
        default=False,
        show_default=True,
        help="Strip load shape profiles from DSS models"
    ),
    click.option(
        "-P",
        "--pv-deployments-dirname",
        help="The output directory name of PV deployments in feeder",
        type=click.STRING,
        default=DEFAULT_PV_DEPLOYMENTS_DIRNAME,
        show_default=True,
    ),
    click.option(
        "-F",
        "--force",
        help="overwrite existing directory",
        is_flag=True,
        default=False,
        show_default=True,
    ),
)


def common_options(func):
    for option in reversed(COMMON_OPTIONS):
        func = option(func)
    return func


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-o",
    "--output",
    default=None,
    show_default=True,
    help="output directory",
)
@click.pass_context
def snapshot(
    ctx,
    hierarchy,
    substations,
    feeders,
    placements,
    samples,
    penetration_levels,
    master_file,
    copy_load_shape_data_files,
    strip_load_shape_profiles,
    pv_deployments_dirname,
    force,
    start,
    output,
):
    """Transform input data for a snapshot simulation"""
    if output is None:
        output = f"snapshot-{hierarchy.value}-models"
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": start,
        "step_resolution": 900,
        "simulation_type": SimulationType.SNAPSHOT,
    }
    SourceTree1Model.transform(
        input_path=input_path,
        output_path=output,
        hierarchy=hierarchy,
        simulation_params=simulation_params,
        simulation_model=SnapshotImpactAnalysisModel,
        substations=substations,
        feeders=feeders,
        placements=placements,
        samples=samples,
        penetration_levels=penetration_levels,
        master_file=master_file,
        copy_load_shape_data_files=copy_load_shape_data_files,
        strip_load_shape_profiles=strip_load_shape_profiles,
        pv_deployments_dirname=pv_deployments_dirname
    )
    print(f"Transformed data from {input_path} to {output} for Snapshot Analysis.")


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-e",
    "--end",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["end_time"],
    show_default=True,
    help="simulation end time",
)
@click.option(
    "-r",
    "--resolution",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["step_resolution"],
    type=int,
    show_default=True,
    help="simulation step resolution in seconds",
)
@click.option(
    "-o",
    "--output",
    default=None,
    show_default=True,
    help="output directory",
)
@click.pass_context
def time_series(
    ctx,
    hierarchy,
    substations,
    feeders,
    placements,
    samples,
    penetration_levels,
    master_file,
    copy_load_shape_data_files,
    strip_load_shape_profiles,
    pv_deployments_dirname,
    force,
    start,
    end,
    resolution,
    output,
):
    """Transform input data for a time series simulation"""
    if output is None:
        output = f"time-series-{hierarchy.value}-models"
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": end,
        "step_resolution": resolution,
        "simulation_type": SimulationType.QSTS,
    }
    SourceTree1Model.transform(
        input_path=input_path,
        output_path=output,
        simulation_params=simulation_params,
        simulation_model=TimeSeriesAnalysisModel,
        substations=substations,
        feeders=feeders,
        placements=placements,
        samples=samples,
        penetration_levels=penetration_levels,
        master_file=master_file,
        hierarchy=hierarchy,
        copy_load_shape_data_files=copy_load_shape_data_files,
        strip_load_shape_profiles=strip_load_shape_profiles,
        pv_deployments_dirname=pv_deployments_dirname
    )
    print(
        f"Transformed data from {input_path} to {output} for TimeSeries Analysis."
    )


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_UPGRADE_COST_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-o",
    "--output",
    default=None,
    show_default=True,
    help="output directory"
)
@click.pass_context
def upgrade(
    ctx,
    hierarchy,
    substations,
    feeders,
    placements,
    samples,
    penetration_levels,
    master_file,
    copy_load_shape_data_files,
    strip_load_shape_profiles,
    pv_deployments_dirname,
    force,
    start,
    output
):
    """Transform input data for an automated upgrade simulation"""
    if output is None:
        output = f"upgrade-{hierarchy.value}-models"
    
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)

    simulation_params = {
        "start_time": start,
        "end_time": start,
        "step_resolution": 900,
        "simulation_type": SimulationType.SNAPSHOT,
    }
    SourceTree1Model.transform(
        input_path=input_path,
        output_path=output,
        hierarchy=hierarchy,
        simulation_params=simulation_params,
        simulation_model=UpgradeCostAnalysisModel,
        substations=substations,
        feeders=feeders,
        placements=placements,
        samples=samples,
        penetration_levels=penetration_levels,
        master_file=master_file,
        copy_load_shape_data_files=copy_load_shape_data_files,
        strip_load_shape_profiles=strip_load_shape_profiles,
        pv_deployments_dirname=pv_deployments_dirname
    )
    print(f"Transformed data from {input_path} to {output} for UpgradeCostAnalysis.")


class SourceTree1Model(BaseOpenDssModel):
    """OpenDSS Model for Source Tree 1"""

    DEPLOYMENT_FILE = "PVSystems.dss"
    TRANSFORM_SUBCOMMANDS = {
        "snapshot": snapshot,
        "time-series": time_series,
        "upgrade": upgrade,
    }

    def __init__(self, data):
        data = copy.deepcopy(data)
        self._path = data.pop("path")
        self._substation = data.pop("substation")
        self._feeder = data.pop("feeder")
        self._master_file = data.pop("master", "Master_noPV.dss")
        self._placement = data.pop("placement")
        self._sample = data.pop("sample")
        self._penetration_level = data.pop("penetration_level")
        self._loadshape_directory = data.pop("loadshape_directory")
        self._opendss_directory = data.pop("opendss_directory")
        self._pv_locations = data.pop("pv_locations")
        self._pydss_controllers = data.pop("pydss_controllers")
        self._metadata_directory = data.pop("metadata_directory", None)
        if data.pop("is_base_case"):
            self._name = self.make_feeder_base_case_name(
                self._substation,
                self._feeder,
            )
        else:
            self._name = self.make_name(
                self._substation,
                self._feeder,
                self._placement,
                self._sample,
                self._penetration_level,
            )
        data.pop("deployment_file")
        assert not data, str(data)

    @staticmethod
    def get_transform_subcommand(name):
        if name not in SourceTree1Model.TRANSFORM_SUBCOMMANDS:
            raise InvalidParameter(f"{name} is not supported")
        return SourceTree1Model.TRANSFORM_SUBCOMMANDS[name]

    @staticmethod
    def list_transform_subcommands():
        return sorted(list(SourceTree1Model.TRANSFORM_SUBCOMMANDS.keys()))

    @property
    def substation(self):
        return self._substation

    @property
    def feeder(self):
        return self._feeder

    @property
    def dc_ac_ratio(self):
        return 1.15

    @property
    def loadshape_directory(self):
        return self._loadshape_directory

    @property
    def opendss_directory(self):
        return self._opendss_directory

    @property
    def master_file(self):
        return os.path.join(self._opendss_directory, self._master_file)

    @property
    def metadata_directory(self):
        return self._metadata_directory

    @property
    def name(self):
        return self._name

    @property
    def pv_locations(self):
        return self._pv_locations

    @property
    def pydss_controllers(self):
        return self._pydss_controllers

    @staticmethod
    def get_transform_defaults():
        return {
            "substations": "all",
            "feeders": "all",
            "placements": "all",
            "samples": "all",
            "penetration_levels": "all",
            "hierarchy": "feeder",
            "master_file": "Master.dss",
            "force": False
        }

    @classmethod
    def transform(
        cls,
        input_path,
        output_path,
        hierarchy,
        simulation_model,
        simulation_params,
        substations=("all",),
        feeders=("all",),
        placements=("all",),
        samples=("all",),
        penetration_levels=("all",),
        master_file="Master.dss",
        copy_load_shape_data_files=False,
        strip_load_shape_profiles=False,
        pv_deployments_dirname=DEFAULT_PV_DEPLOYMENTS_DIRNAME
    ):
        inputs = SourceTree1ModelInputs(input_path, pv_deployments_dirname)

        if substations == ("all",):
            substations = inputs.list_substations()
        if placements == ("all",):
            placements = inputs.list_placements()
        elif not isinstance(placements[0], float):
            placements = [Placement(x) for x in placements]

        if hierarchy == SimulationHierarchy.FEEDER:
            func = cls._transform_by_feeder
        elif hierarchy == SimulationHierarchy.SUBSTATION:
            func = cls._transform_by_substation
        else:
            assert False

        os.makedirs(output_path, exist_ok=True)
        func(
            inputs,
            input_path,
            output_path,
            simulation_model,
            simulation_params,
            substations,
            feeders,
            placements,
            samples,
            penetration_levels,
            master_file,
            copy_load_shape_data_files,
            strip_load_shape_profiles
        )

    @classmethod
    def _transform_by_feeder(
        cls,
        inputs,
        input_path,
        output_path,
        simulation_model,
        simulation_params,
        substations,
        feeders,
        placements,
        samples,
        penetration_levels,
        master_file,
        copy_load_shape_data_files,
        strip_load_shape_profiles
    ):
        config = []
        base_cases = set()
        for substation, placement in itertools.product(substations, placements):
            for feeder in inputs.list_feeders(substation):
                if feeders != ("all",) and feeder not in feeders:
                    continue
                base_case_name = SourceTree1Model.make_feeder_base_case_name(substation, feeder)
                if base_case_name not in base_cases and issubclass(simulation_model, ImpactAnalysisBaseModel):
                    data = {
                        "path": input_path,
                        "substation": substation,
                        "feeder": feeder,
                        "master": master_file,
                        "placement": None,
                        "sample": None,
                        "penetration_level": None,
                        "deployment_file": None,
                        "loadshape_directory": None,
                        "opendss_directory": inputs.get_opendss_directory(
                            substation, feeder
                        ),
                        "metadata_directory": inputs.get_metadata_directory(substation, feeder),
                        "pv_locations": [],
                        "pydss_controllers": None,
                        "is_base_case": True,
                    }
                    model = cls(data)
                    path = os.path.join(output_path, substation, feeder)
                    out_deployment = model.create_base_case(
                        base_case_name,
                        path,
                        copy_load_shape_data_files=copy_load_shape_data_files,
                    )
                    item = {
                        "deployment": out_deployment,
                        "simulation": simulation_params,
                        "name": base_case_name,
                        "model_type": simulation_model.__name__,
                        "job_order": 0,
                        "is_base_case": True,
                    }
                    config.append(simulation_model.validate(item).dict())
                    base_cases.add(base_case_name)
                key = inputs.create_key(substation, feeder, placement)
                if samples == ("all",):
                    _samples = inputs.list_samples(key)
                else:
                    _samples = [int(x) for x in samples]
                for sample in _samples:
                    if penetration_levels == ("all",):
                        levels = inputs.list_penetration_levels(key, sample)
                    else:
                        levels = [int(x) for x in penetration_levels]
                    pv_configs = inputs.list_pv_configs(
                        substation, feeder, placement, sample
                    )
                    for level in levels:
                        deployment_file = inputs.get_deployment_file(key, sample, level)
                        pydss_controller, pv_profiles = get_pydss_controller_and_profiles(pv_configs)
                        data = {
                            "path": input_path,
                            "substation": substation,
                            "feeder": feeder,
                            "master": master_file,
                            "placement": placement.value,
                            "sample": sample,
                            "penetration_level": level,
                            "deployment_file": deployment_file,
                            "loadshape_directory": None,
                            "opendss_directory": inputs.get_opendss_directory(
                                substation, feeder
                            ),
                            "metadata_directory": inputs.get_metadata_directory(substation, feeder),
                            "pv_locations": [deployment_file],
                            "pydss_controllers": pydss_controller,
                            "is_base_case": False,
                        }
                        model = cls(data)
                        path = os.path.join(output_path, substation, feeder)
                        out_deployment = model.create_deployment(
                            model.name,
                            path,
                            pv_profile=pv_profiles,
                            hierarchy=SimulationHierarchy.FEEDER,
                            copy_load_shape_data_files=copy_load_shape_data_files,
                            strip_load_shape_profiles=strip_load_shape_profiles
                        )
                        out_deployment.project_data["placement"] = placement
                        out_deployment.project_data["sample"] = sample
                        out_deployment.project_data["penetration_level"] = level
                        item = {
                            "deployment": out_deployment,
                            "simulation": simulation_params,
                            "name": model.name,
                            "model_type": simulation_model.__name__,
                            "job_order": level
                        }
                        if issubclass(simulation_model, ImpactAnalysisBaseModel):
                            item["base_case"] = base_case_name
                        config.append(simulation_model.validate(item).dict())

        filename = os.path.join(output_path, SOURCE_CONFIGURATION_FILENAME)
        with open(filename, "w") as f_out:
            json.dump(config, f_out, indent=2, cls=ExtendedJSONEncoder)
        logger.info("Wrote config to %s", filename)

    @classmethod
    def _transform_by_substation(
            cls,
            inputs,
            input_path,
            output_path,
            simulation_model,
            simulation_params,
            substations,
            feeders,
            placements,
            samples,
            penetration_levels,
            master_file,
            copy_load_shape_data_files,
            strip_load_shape_profiles
    ):
        config = []
        deployment_files_by_key = defaultdict(list)
        base_case_names = set()
        for substation, placement in itertools.product(substations, placements):
            if issubclass(simulation_model, ImpactAnalysisBaseModel):
                base_case_name = cls.make_substation_base_case_name(substation)
                if base_case_name not in base_case_names:
                    data = {
                        "path": input_path,
                        "substation": substation,
                        "feeder": None,
                        "master": master_file,
                        "placement": None,
                        "sample": None,
                        "penetration_level": None,
                        "deployment_file": None,
                        "loadshape_directory": None,
                        "opendss_directory": inputs.get_substation_opendss_directory(
                            substation,
                        ),
                        "pv_locations": [],
                        "pydss_controllers": None,
                        "is_base_case": True,
                    }
                    model = cls(data)
                    path = os.path.join(output_path, substation)
                    out_deployment = model.create_substation_base_case(
                        base_case_name,
                        path,
                        copy_load_shape_data_files=copy_load_shape_data_files,
                    )
                    item = {
                        "deployment": out_deployment,
                        "simulation": simulation_params,
                        "name": base_case_name,
                        "model_type": simulation_model.__name__,
                        "job_order": 0,
                        "is_base_case": True,
                    }
                    config.append(simulation_model.validate(item).dict())
                    base_case_names.add(base_case_name)
                    fix_substation_master_file(Path(output_path) / substation / master_file)
            for feeder in inputs.list_feeders(substation):
                if feeders != ("all",) and feeder not in feeders:
                    continue
                key = inputs.create_key(substation, feeder, placement)
                if samples == ("all",):
                    _samples = inputs.list_samples(key)
                else:
                    _samples = [int(x) for x in samples]
                for sample in _samples:
                    if penetration_levels == ("all",):
                        levels = inputs.list_penetration_levels(key, sample)
                    else:
                        levels = [int(x) for x in penetration_levels]
                    for level in levels:
                        substation_key = SubstationKey(substation, placement, sample, level)
                        job_deployment_file = inputs.get_deployment_file(key, sample, level)
                        pv_configs = inputs.list_pv_configs(
                            substation, feeder, placement, sample
                        )
                        pydss_controller, pv_profiles = get_pydss_controller_and_profiles(pv_configs)
                        data = {
                            "path": input_path,
                            "substation": substation,
                            "feeder": "None",
                            "master": master_file,
                            "placement": placement.value,
                            "sample": sample,
                            "penetration_level": level,
                            "deployment_file": job_deployment_file,
                            "loadshape_directory": None,
                            "opendss_directory": inputs.get_opendss_directory(substation, feeder),
                            "pv_locations": [job_deployment_file],
                            "pydss_controllers": pydss_controller,
                            "is_base_case": False,
                        }
                        model = cls(data)
                        path = os.path.join(output_path, substation, feeder)
                        out_deployment = model.create_deployment(
                            model.name,
                            path,
                            pv_profile=pv_profiles,
                            hierarchy=SimulationHierarchy.SUBSTATION,
                            copy_load_shape_data_files=copy_load_shape_data_files,
                            strip_load_shape_profiles=strip_load_shape_profiles
                        )
                        out_deployment.project_data["placement"] = placement
                        out_deployment.project_data["sample"] = sample
                        out_deployment.project_data["penetration_level"] = level
                        out_job_deployment_file = out_deployment.deployment_file
                        if not deployment_files_by_key[substation_key]:
                            # Need to create a job at the substation level.
                            # All params except for the file and name are the same as the job.
                            out_deployment.deployment_file = str(substation_key_to_dss_filename(output_path, substation_key))
                            item = {
                                "deployment": out_deployment,
                                "simulation": simulation_params,
                                "name": cls.make_substation_job_name(substation_key),
                                "model_type": simulation_model.__name__,
                                "job_order": level,
                            }
                            if issubclass(simulation_model, ImpactAnalysisBaseModel):
                                item["base_case"] = base_case_name
                            config.append(simulation_model.validate(item).dict())
                        deployment_files_by_key[substation_key].append(out_job_deployment_file)

        for substation_key, deployment_files in deployment_files_by_key.items():
            make_substation_pv_deployments(output_path, substation_key, deployment_files)

        filename = os.path.join(output_path, SOURCE_CONFIGURATION_FILENAME)
        with open(filename, "w") as f_out:
            json.dump(config, f_out, indent=2, cls=ExtendedJSONEncoder)
        logger.info("Wrote config to %s", filename)

    @staticmethod
    def make_name(substation, feeder, placement, sample, penetration_level):
        fields = (substation, feeder, placement, sample, penetration_level)
        return "__".join([str(x) for x in fields])

    @staticmethod
    def make_substation_job_name(key: SubstationKey):
        return "__".join(
            (
                key.substation,
                key.placement.value,
                str(key.sample),
                str(key.penetration_level),
            )
        )

    @staticmethod
    def make_feeder_base_case_name(substation, feeder):
        fields = (substation, feeder, -1, -1, -1)
        return "__".join([str(x) for x in fields])

    @staticmethod
    def make_substation_base_case_name(substation):
        fields = (substation, -1, -1, -1)
        return "__".join([str(x) for x in fields])


def make_substation_pv_deployments(output_path, key, deployment_files):
    filename = substation_key_to_dss_filename(output_path, key)
    with open(filename, "w") as f_out:
        f_out.write(f"Redirect ../Master.dss\n\n")
        for deployment_file in deployment_files:
            # We have the option of either redirecting to each file or include the content
            # of each file.
            # I'm choosing the latter because PyDSS has code to parse all PVSystems from
            # one file in order to construct PvController.toml files, and so this makes that work.
            # We could make that recurse.
            # It might also be helpful for debugging and analysis to see the PVSystems in one
            # file.
            f_out.write(Path(deployment_file).read_text())
            os.remove(deployment_file)
        f_out.write("\nSolve\n")
    logger.info("Wrote substation-level deployment files to %s", filename)


def substation_key_to_dss_filename(output_path, key):
    path = Path(output_path) / key.substation / "PVDeployments"
    os.makedirs(path, exist_ok=True)
    return path / (SourceTree1Model.make_substation_job_name(key) + ".dss")


def get_pydss_controller_and_profiles(pv_configs):
    pydss_controllers = set()
    pv_profiles = {}
    pydss_controller = None
    for pv_config in pv_configs:
        if pv_config["pydss_controller"] is None:
            ctrl = None
        else:
            ctrl = (pv_config["pydss_controller"]["controller_type"],
                    pv_config["pydss_controller"]["name"])
            if ctrl[1] != "pf1" and ctrl not in pydss_controllers:
                pydss_controller = PyDSSControllerModel.validate(
                    pv_config["pydss_controller"]
                )
                pydss_controllers.add(ctrl)
        pv_profiles[pv_config["name"]] = pv_config.get("pv_profile")

    if len(pydss_controllers) > 1:
        raise Exception(
            f"only 1 pydss controller is currently supported: {pydss_controllers}"
        )

    #assert pydss_controller is not None
    return pydss_controller, pv_profiles


def fix_substation_master_file(filename):
    # These master files have "Redirect <substation>--<feeder>".
    # For better or worse, DISCO has already removed substation and then added
    # "/OpenDSS", so we have to patch the references here.
    # Use "edirect" instead of "Redirect" to avoid case-insensitive checks.
    regex = re.compile(r"edirect \w+--(\w+)")

    def remove_substation(match):
        return "edirect " + match.group(1) + "/OpenDSS"

    with fileinput.input(files=[filename], inplace=True) as f_in:
        for line in f_in:
            line = re.sub(regex, remove_substation, line)
            print(line, end="")



"""Contains functionality to configure PyDss simulations."""

import logging
from collections import defaultdict

from jade.exceptions import InvalidParameter, InvalidConfiguration
from jade.extensions.generic_command.generic_command_parameters import GenericCommandParameters
from jade.utils.utils import load_data

from PyDSS.common import ControllerType
from PyDSS.reports.pv_reports import CONTROL_MODE_SCENARIO

from disco.distribution.deployment_parameters import DeploymentParameters
from disco.enums import SimulationHierarchy, SimulationType
from disco.pydss.pydss_configuration_base import PyDssConfigurationBase
from disco.pydss.common import ConfigType
from disco.extensions.pydss_simulation.pydss_inputs import PyDssInputs

logger = logging.getLogger(__name__)


class PyDssConfiguration(PyDssConfigurationBase):
    """Represents the configuration options for a PyDSS simulation."""

    def __init__(self, **kwargs):
        """Constructs PyDssConfiguration."""
        super(PyDssConfiguration, self).__init__(**kwargs)
        self._scenario_names = []

    def add_hosting_capacity_job(self, simulation_type, impact_analysis_jobs):
        """Add post-processing jobs for hosting capacity.

        Parameters
        ----------
        simulation_type : SimulationType
        impact_analysis_jobs : list
            Names of impact analysis jobs

        """
        if simulation_type == SimulationType.SNAPSHOT:
            command = "compute-snapshot-hosting-capacity"
        elif simulation_type == SimulationType.QSTS:
            command = "compute-time-series-hosting-capacity"
        else:
            assert False, simulation_type

        cmd = f"disco-internal {command}"
        job = GenericCommandParameters(
            command=cmd,
            job_id=command,
            blocked_by=impact_analysis_jobs,
            append_output_dir=True,
        )
        self.add_job(job)

    def add_impact_analysis_jobs(self, simulation_type):
        """Add post-processing jobs for impact analysis.

        Parameters
        ----------
        simulation_type : SimulationType

        Returns
        -------
        list
            Names of impact analysis jobs

        """
        if simulation_type == SimulationType.SNAPSHOT:
            command = "compute-snapshot-impact-analysis"
        elif simulation_type == SimulationType.QSTS:
            command = "compute-time-series-impact-analysis"
        else:
            assert False, simulation_type

        feeders = defaultdict(list)
        for job in self.iter_jobs():
            feeders[job.model.deployment.feeder].append(job.name)

        job_names = []
        for feeder, blocking_jobs in feeders.items():
            cmd = f"disco-internal {command} {feeder}"
            name = f"{feeder}-{command}"
            job = GenericCommandParameters(
                command=cmd,
                job_id=name,
                blocked_by=blocking_jobs,
                append_output_dir=True,
            )
            self.add_job(job)
            job_names.append(name)

        return job_names

    @classmethod
    def auto_config(cls, inputs, simulation_config=None, scenarios=None,
                    order_by_penetration=False, estimated_run_minutes=None,
                    dc_ac_ratio=None, **kwargs):
        """Create a configuration from all available inputs."""
        if isinstance(inputs, str):
            inputs = PyDssInputs(inputs)
        config = cls(**kwargs)
        for job in inputs.iter_jobs():
            job.estimated_run_minutes = estimated_run_minutes
            if dc_ac_ratio is not None:
                job.model.deployment.dc_ac_ratio = dc_ac_ratio
            config.add_job(job)

        if simulation_config is None:
            simulation_config = config.get_default_pydss_simulation_config()
        config.set_pydss_config(ConfigType.SIMULATION_CONFIG, simulation_config)

        if scenarios is None:
            scenarios = [PyDssConfiguration.make_default_pydss_scenario("scenario")]
        config.set_pydss_config(ConfigType.SCENARIOS, scenarios)

        if order_by_penetration:
            config.apply_job_order_by_penetration_level()

        config.check_job_consistency()
        return config

    def apply_job_order_by_penetration_level(self):
        """Assign blocked_by field in each job."""
        keys = set(("placement", "sample", "penetration_level"))
        jobs_to_order = defaultdict(list)
        for job in self.iter_pydss_simulation_jobs():
            if len(keys.intersection(set(job.model.deployment.project_data.keys()))) == 3:
                key = (
                    job.model.deployment.substation,
                    job.model.deployment.feeder,
                    job.model.deployment.project_data["placement"],
                    job.model.deployment.project_data["sample"],
                )
                jobs_to_order[key].append(job)

        # If a job of a lower penetration level fails with a convergence error then
        # all higher jobs are expected to fail. This allows us to cancel those jobs.
        for jobs in jobs_to_order.values():
            if len(jobs) < 2:
                continue
            jobs.sort(key=lambda x: x.model.deployment.project_data["penetration_level"])
            for i, job in enumerate(jobs[1:]):
                # Blocked by the previous job.
                job.add_blocking_job(jobs[i].name)

    def check_job_consistency(self):
        """Check all jobs for consistency.
        
        Raises
        ------
        InvalidConfiguration
            Raised if a rule is violated.

        """
        config_has_pydss_controllers = None
        for job in self.iter_pydss_simulation_jobs(exclude_base_case=True):
            if config_has_pydss_controllers is None:
                config_has_pydss_controllers = job.has_pydss_controllers()
            elif job.has_pydss_controllers() != config_has_pydss_controllers:
                raise InvalidParameter("All jobs must consistently define pydss_controllers.")
        
    def create_from_result(self, job, output_dir):
        cls = self.job_execution_class(job.extension)
        return cls.create(self.pydss_inputs, job, output=output_dir)

    def get_base_case_job(self, feeder):
        """Return the base_case job for the given feeder.

        Parameters
        ----------
        feeder : str

        """
        for job in self.iter_feeder_jobs(feeder):
            if job.model.is_base_case:
                return job

        return InvalidParameter(f"no base case job for feeder={feeder}")

    def get_simulation_hierarchy(self):
        """Return the SimulationHierarchy for the config.

        Returns
        -------
        SimulationHierarchy

        """
        for job in self.iter_pydss_simulation_jobs():
            if not job.model.is_base_case:
                if job.model.deployment.feeder == "None":
                    hierarchy = SimulationHierarchy.SUBSTATION
                else:
                    hierarchy = SimulationHierarchy.FEEDER
                return hierarchy

        raise Exception("Failed to identify SimulationHierarchy")

    def has_pydss_controllers(self):
        """Return True if the jobs have pydss controllers defined."""
        job = next(iter(self.iter_pydss_simulation_jobs(exclude_base_case=True)))
        return job.has_pydss_controllers()

    def iter_feeder_jobs(self, feeder):
        """Return jobs for the given feeder in the config.

        Parameters
        ----------
        feeder : str
            Return jobs with this feeder.

        Yields
        ------
        DeploymentParameters

        """
        for job in self.iter_pydss_simulation_jobs():
            if job.feeder == feeder:
                yield job

    def iter_pydss_simulation_jobs(self, exclude_base_case=False):
        """Return jobs that are pydss_simulation jobs (not post-processing).

        Parameters
        ----------
        exclude_base_case : bool

        Yields
        ------
        DeploymentParameters

        """
        for job in self.iter_jobs():
            if isinstance(job, DeploymentParameters):
                if exclude_base_case and job.model.is_base_case:
                    continue
                yield job

    def list_feeders(self):
        """Return a list of unique feeders in the config.

        Returns
        -------
        list

        """
        return list({x.feeder for x in self.iter_pydss_simulation_jobs()})

    def get_feeder_job(self, feeder, job_name):
        """Return job by name

        Parameters
        ----------
        feeder : str
        job_name : str

        Returns
        -------
        DeploymentParameters

        """
        # TODO: callers should just call this.
        return self.get_job(job_name)

    def update_volt_var_curve(self, volt_var_curve: str):
        """Update the volt-var curve for all jobs configured with PV controls."""
        count = 0
        for job in self.iter_pydss_simulation_jobs(exclude_base_case=True):
            pydss_controllers = job.model.deployment.pydss_controllers
            if pydss_controllers is not None:
                if isinstance(pydss_controllers, list):
                    for controller in pydss_controllers:
                        if controller.controller_type == ControllerType.PV_CONTROLLER:
                            controller.name = volt_var_curve
                            count += 1
                elif pydss_controllers.controller_type == ControllerType.PV_CONTROLLER:
                    pydss_controllers.name = volt_var_curve
                    count += 1

        logger.info("Updated %s jobs with volt-var curve %s", count, volt_var_curve)



import logging

import click
from jade.common import CONFIG_FILE
from jade.loggers import setup_logging
from jade.utils.utils import load_data

from disco.extensions.upgrade_simulation.upgrade_inputs import UpgradeInputs
from disco.extensions.upgrade_simulation.upgrade_configuration import (
    UpgradeConfiguration,
    DEFAULT_UPGRADE_COST_DB_FILE,
    DEFAULT_UPGRADE_PARAMS_FILE
)
from disco.models.upgrade_cost_analysis_generic_input_model import (
    ThermalUpgradeParamsModel,
    VoltageUpgradeParamsModel
)

logger = logging.getLogger(__name__)


@click.command()
@click.argument("inputs")
@click.option(
    "-d", "--cost-database",
    type=click.Path(exists=True),
    default=DEFAULT_UPGRADE_COST_DB_FILE,
    show_default=True,
    help="The unit cost database spreadsheet."
)
@click.option(
    "-p", "--params-file",
    type=click.Path(exists=True),
    required=False,
    default=DEFAULT_UPGRADE_PARAMS_FILE,
    show_default=True,
    help="Upgrade parameters file."
)
@click.option(
    "-c", "--config-file",
    type=click.Path(),
    default=CONFIG_FILE,
    show_default=True,
    help="JADE config file to create"
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Enable debug logging."
)
def upgrade(
    inputs,
    cost_database,
    params_file,
    config_file,
    verbose=False
):
    """Create JADE configuration for upgrade simulations"""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])

    params = load_data(params_file)
    thermal_upgrade_params = ThermalUpgradeParamsModel(**params["thermal_upgrade_params"]).dict()
    voltage_upgrade_params = VoltageUpgradeParamsModel(**params["voltage_upgrade_params"]).dict()
    upgrade_simulation_params = params["upgrade_simulation_params"]

    job_global_config = {
        "upgrade_cost_database": cost_database,
        "thermal_upgrade_params": thermal_upgrade_params,
        "voltage_upgrade_params": voltage_upgrade_params,
        "upgrade_simulation_params": upgrade_simulation_params
    }

    inputs = UpgradeInputs(inputs)
    config = UpgradeConfiguration.auto_config(
        inputs=inputs,
        job_global_config=job_global_config
    )
    config.dump(filename=config_file)
    print(f"Created {config_file} for upgrade cost analysis.")



import os

import click

from jade.utils.utils import load_data, dump_data
from PyDSS.common import SnapshotTimePointSelectionMode
import disco
from disco.postprocess.hosting_capacity import compute_hc


def _handle_snapshot_time_point(_, __, val):
    return SnapshotTimePointSelectionMode(val)


@click.command()
@click.argument("output_dir")
@click.option(
    "--metric-class",
    type=click.Choice(["thermal", "voltage"], case_sensitive=True),
    multiple=True,
    default=("thermal", "voltage"),
    show_default=True,
    help="Choose the metric class"
)
@click.option(
    "--on",
    type=click.STRING,
    default=("all", ),
    multiple=True,
    help="The metric included for query, default 'all'"
)
@click.option(
    "--scenario",
    type=click.Choice(["scenario", "control_mode", "pf1", "derms"], case_sensitive=True),
    default="scenario",
    show_default=True,
    help="Choose the PyDSS scenario"
)
@click.option(
    "--node-type",
    type=click.Choice(["primaries", "secondaries"], case_sensitive=True),
    multiple=True,
    default=("primaries", "secondaries"),
    show_default=True,
    help="Choose the node type."
)
@click.option(
    "--hc-thresholds",
    type=click.Path(),
    default=os.path.join(
        os.path.dirname(getattr(disco, "__path__")[0]),
        "disco",
        "postprocess",
        "config",
        "hc_thresholds.toml",
    ),
    show_default=True,
    help="The thresholds for filtering metrics"
)
@click.option(
    "--time-point",
    type=click.Choice([x.value for x in SnapshotTimePointSelectionMode]),
    default=SnapshotTimePointSelectionMode.NONE.value,
    show_default=True,
    callback=_handle_snapshot_time_point,
    help="Time point selection type. Only applicable to snapshot simulations."
)
def compute_hosting_capacity(output_dir, metric_class, on, scenario, node_type, hc_thresholds, time_point):
    """Compute the hosting capacity"""

    if "JADE_PIPELINE_OUTPUT_DIR" in os.environ and "$JADE_PIPELINE_OUTPUT_DIR" in output_dir:
        pipeline_output = os.environ["JADE_PIPELINE_OUTPUT_DIR"]
        output_dir = output_dir.replace("$JADE_PIPELINE_OUTPUT_DIR", pipeline_output)

    if on == ("all", ):
        on = "all"

    hc_summary, overall_hc, query_list = compute_hc(
        result_path=output_dir,
        thresholds=load_data(hc_thresholds),
        metric_classes=metric_class,
        scenario=scenario,
        on=on,
        node_types=node_type
    )

    print("Query List:", query_list)

    # TODO DISCO-315: handle the time_point value for snapshots.
    suffix = "" if time_point == SnapshotTimePointSelectionMode.NONE else f"__{time_point.value}"
    hc_summary_file = os.path.join(output_dir, f"hosting_capacity_summary__{scenario}{suffix}.json")
    dump_data(hc_summary, hc_summary_file, indent=2)
    print(f"Hosting Capacity Summary data is dumped - {hc_summary_file}")

    overall_hc_file = os.path.join(output_dir, f"hosting_capacity_overall__{scenario}{suffix}.json")
    dump_data(overall_hc, overall_hc_file, indent=2)
    print(f"Hosting Capacity Overall data is dumped - {overall_hc_file}")
